#    test_merge_upstream.py -- Testsuite for builddeb's upstream merging.
#    Copyright (C) 2007 James Westby <jw+debian@jameswestby.net>
#
#    This file is part of bzr-builddeb.
#
#    bzr-builddeb is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    bzr-builddeb is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with bzr-builddeb; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#

from debian.changelog import Changelog, Version

from ....tests import (
    TestCaseWithTransport,
    )

from ..merge_upstream import (
    changelog_add_new_version,
    )


class ChangelogAddNewVersionTests(TestCaseWithTransport):

    def test_add_new(self):
        tree = self.make_branch_and_tree(".")
        tree.lock_write()
        self.addCleanup(tree.unlock)
        tree.mkdir("debian")
        changelog_add_new_version(tree, '', "1.0", "sid", None, "somepkg")
        # changelog_add_new_version will version the changelog if it was
        # created
        with open('debian/changelog', 'rb') as f:
            cl = Changelog(f)
        self.assertEqual(cl._blocks[0].package, "somepkg")
        self.assertEqual(cl._blocks[0].distributions, "UNRELEASED")
        self.assertEqual(cl._blocks[0].version, Version("1.0-1"))
        self.assertEqual(
            [], list(tree.filter_unversioned_files(["debian/changelog"])))



#    tags.py -- Providers of upstream source - tag names
#    Copyright (C) 2016-2020 Jelmer Vernooij <jelmer@debian.org>
#
#    This file is part of bzr-builddeb.
#
#    bzr-builddeb is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    bzr-builddeb is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with bzr-builddeb; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

from itertools import islice
from typing import Optional

from ....repository import Repository
from ....revision import Revision


from debmutate.versions import mangle_version_for_git


def possible_upstream_tag_names(package: Optional[str], version: str,
                                component: Optional[str] = None,
                                try_hard=True):
    tags = []
    if component is None:
        # compatibility with git-buildpackage
        tags.append("upstream/%s" % version)
        tags.append("upstream-%s" % version)
        manipulated = 'upstream/%s' % mangle_version_for_git(version)
        if manipulated not in tags:
            tags.append(manipulated)
        # compatibility with svn-buildpackage
        tags.append("upstream_%s" % version)

        if try_hard:
            # common upstream names
            tags.append("%s" % version)
            tags.append("v%s" % version)
            if '~' not in str(version) and '+' not in str(version):
                tags.append("release-%s" % version)
                tags.append("v%s-release" % version)
            if package:
                tags.append("{}-{}".format(package, version))
            tags.append("v/%s" % version)
            tags.append("v.%s" % version)
    else:
        tags.append('upstream-{}/{}'.format(version, component))
        tags.append('upstream/{}/{}'.format(
            mangle_version_for_git(version), component))
    return tags


def is_upstream_tag(tag):
    """Return true if tag is an upstream tag.

    :param tag: The string name of the tag.
    :return: True if the tag name is one generated by upstream tag operations.
    """
    return (tag.startswith('upstream-') or tag.startswith('upstream/') or
            tag.startswith('upstream_'))


def upstream_tag_version(tag):
    """Return the upstream version portion of an upstream tag name.

    :param tag: The string name of the tag.
    :return: tuple with version portion of the tag and component name
    """
    assert is_upstream_tag(tag), "Not an upstream tag: %s" % tag
    if tag.startswith('upstream/'):
        tag = tag[len('upstream/'):]
    elif tag.startswith('upstream_'):
        tag = tag[len('upstream_'):]
    elif tag.startswith('upstream-'):
        tag = tag[len('upstream-'):]
        if tag.startswith('debian-'):
            tag = tag[len('debian-'):]
        elif tag.startswith('ubuntu-'):
            tag = tag[len('ubuntu-'):]
    tag = tag.replace('_', '~')
    if '/' not in tag:
        return (None, tag)
    (version, component) = tag.rsplit('/', 1)
    if component == "":
        component = None
    return (component, version)


def _rev_is_upstream_import(
        revision: Revision, package: Optional[str], version: str):
    possible_messages = [
    ]
    if package is not None:
        possible_messages.extend([
            'Import {}_{}'.format(package, version),
            'import {}_{}'.format(package, version),
            'import {}-{}'.format(package.replace('-', '_'), version),
            '{}-{}'.format(package, version),
        ])
    possible_messages.extend([
        'Imported upstream version %s' % version,
        'Import upstream version %s' % version,
        'New upstream version %s' % version,
        'New upstream version v%s' % version,
        ])
    for possible_message in possible_messages:
        if revision.message.lower().startswith(possible_message.lower()):
            return True
    return False


def _rev_is_upstream_merge(
        revision: Revision, package: Optional[str], version: str) -> bool:
    if revision.message.lower().startswith(
            ("Merge tag 'v%s' into debian/" % version).lower()):
        return True
    if package is not None and revision.message.lower().startswith(
            ("Merge tag '{}-{}' into ".format(package, version)).lower()):
        return True
    return False


def upstream_version_tag_start_revids(
        tag_dict, package: Optional[str], version: str):
    """Find Debian tags related to a particular upstream version.

    This can be used by search_for_upstream_version
    """
    candidate_tag_start = [
        'debian/%s-' % mangle_version_for_git(version),
        'debian-%s' % version,
        # Epochs are sometimes replaced by underscores, rather than by %,
        # as DEP-14 suggests.
        'debian/%s-' % mangle_version_for_git(version.replace(':', '_')),
        # Haskell repo style
        "{}_v{}".format(package, version),
        ]
    if package:
        candidate_tag_start.append('debian-{}-{}'.format(package, version))
    for tag_name, revid in tag_dict.items():
        if any([tag_name.startswith(tag_start)
                for tag_start in candidate_tag_start]):
            yield (tag_name, revid)


def search_for_upstream_version(
        repository: Repository, start_revids, package: Optional[str],
        version: str, component: Optional[str] = None,
        md5: Optional[str] = None,
        scan_depth=None):
    """Find possible upstream revisions that don't have appropriate tags."""
    todo = []
    graph = repository.get_graph()
    for revid, parents in islice(
            graph.iter_ancestry(start_revids), scan_depth):
        todo.append(revid)
    for revid, rev in repository.iter_revisions(todo):
        if rev is None:
            continue
        if _rev_is_upstream_import(rev, package, version):
            return revid

    # Try again, but this time search for merge revisions
    for revid, rev in repository.iter_revisions(todo):
        if rev is None:
            continue
        if _rev_is_upstream_merge(rev, package, version):
            return rev.parent_ids[1]
    return None



#    merge_package.py -- The plugin for bzr
#    Copyright (C) 2009 Canonical Ltd.
#    Copyright (C) 2022 Jelmer Vernooĳ
#
#    :Author: Muharem Hrnjadovic <muharem@ubuntu.com>
#
#    This file is part of bzr-builddeb.
#
#    bzr-builddeb is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    bzr-builddeb is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with bzr-builddeb; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#

import json
import logging
import os
import re
import sys
import tempfile

from debian.changelog import Version, Changelog
from debian.deb822 import Deb822

from breezy.errors import (
    BzrError,
    ConflictsInTree,
    DivergedBranches,
    NoSuchTag,
    UnrelatedBranches,
    )
from breezy.transport import NoSuchFile
from breezy.workingtree import (
    PointlessMerge,
    )

from debian.changelog import format_date, get_maintainer
from debmutate.changelog import (
    ChangelogEditor, changeblock_add_line, increment_version)
from debmutate.reformatting import GeneratedFile

from .changelog import debcommit
from .cmds import _build_helper
from .directory import vcs_field_to_bzr_url_converters
from .errors import (
    MultipleUpstreamTarballsNotSupported,
)
from .import_dsc import DistributionBranch
from .util import (
    dput_changes,
    debsign,
    find_changelog,
    MissingChangelogError,
)


class ChangelogGeneratedFile(Exception):
    """The changelog file is generated."""

    def __init__(self, path, template_path, template_type):
        self.path = path
        self.template_path = template_path
        self.template_type = template_type


class SharedUpstreamConflictsWithTargetPackaging(BzrError):

    _fmt = ('The upstream branches for the merge source and target have '
            'diverged. Unfortunately, the attempt to fix this problem '
            'resulted in conflicts. Please resolve these, commit and '
            're-run the "%(cmd)s" command to finish. '
            'Alternatively, until you commit you can use "bzr revert" to '
            'restore the state of the unmerged branch.')

    def __init__(self, cmd):
        self.cmd = cmd


def _upstream_version_data(branch, revid):
    """Most recent upstream versions/revision IDs of the merge source/target.

    Please note: both packaging branches must have been read-locked
    beforehand.

    :param branch: The merge branch.
    :param revid: The revision in the branch to consider
    :param tree: Optional tree for the revision
    """
    db = DistributionBranch(branch, branch)
    tree = branch.repository.revision_tree(revid)
    changelog, _ignore = find_changelog(tree, '', False)
    uver = changelog.version.upstream_version
    upstream_revids = db.pristine_upstream_source.version_as_revisions(
        None, uver)
    if list(upstream_revids.keys()) != [None]:
        raise MultipleUpstreamTarballsNotSupported()
    upstream_revid, upstream_subpath = upstream_revids[None]
    return (Version(uver), upstream_revid, upstream_subpath)


def fix_ancestry_as_needed(tree, source, source_revid=None):
    r"""Manipulate the merge target's ancestry to avoid upstream conflicts.

    Merging J->I given the following ancestry tree is likely to result in
    upstream merge conflicts:

    debian-upstream                 ,------------------H
                       A-----------B                    \
    ubuntu-upstream     \           \`-------G           \
                         \           \        \           \
    debian-packaging      \ ,---------D--------\-----------J
                           C           \        \
    ubuntu-packaging        `----E------F--------I

    Here there was a new upstream release (G) that Ubuntu packaged (I), and
    then another one that Debian packaged, skipping G, at H and J.

    Now, the way to solve this is to introduce the missing link.

    debian-upstream                 ,------------------H------.
                       A-----------B                    \      \
    ubuntu-upstream     \           \`-------G-----------\------K
                         \           \        \           \
    debian-packaging      \ ,---------D--------\-----------J
                           C           \        \
    ubuntu-packaging        `----E------F--------I

    at K, which isn't a real merge, as we just use the tree from H, but add
    G as a parent and then we merge that in to Ubuntu.

    debian-upstream                 ,------------------H------.
                       A-----------B                    \      \
    ubuntu-upstream     \           \`-------G-----------\------K
                         \           \        \           \      \
    debian-packaging      \ ,---------D--------\-----------J      \
                           C           \        \                  \
    ubuntu-packaging        `----E------F--------I------------------L

    At this point we can merge J->L to merge the Debian and Ubuntu changes.

    :param tree: The `WorkingTree` of the merge target branch.
    :param source: The merge source (packaging) branch.
    """
    upstreams_diverged = False
    t_upstream_reverted = False
    target = tree.branch

    with source.lock_read():
        if source_revid is None:
            source_revid = source.last_revision()
        with tree.lock_write():
            # "Unpack" the upstream versions and revision ids for the merge
            # source and target branch respectively.
            (us_ver, us_revid, us_subpath) = _upstream_version_data(
                source, source_revid)
            (ut_ver, ut_revid, ut_subpath) = _upstream_version_data(
                target, target.last_revision())

            if us_subpath or ut_subpath:
                raise Exception("subpaths not yet supported")

            # Did the upstream branches of the merge source/target diverge?
            graph = source.repository.get_graph(target.repository)
            upstreams_diverged = (len(graph.heads([us_revid, ut_revid])) > 1)

            # No, we're done!
            if not upstreams_diverged:
                return (upstreams_diverged, t_upstream_reverted)

            # Instantiate a `DistributionBranch` object for the merge target
            # (packaging) branch.
            db = DistributionBranch(tree.branch, tree.branch)
            with tempfile.TemporaryDirectory(
                    dir=os.path.join(tree.basedir, '..')) as tempdir:
                # Extract the merge target's upstream tree into a temporary
                # directory.
                db.extract_upstream_tree(
                    {None: (ut_revid, ut_subpath)}, tempdir)
                tmp_target_utree = db.pristine_upstream_tree

                # Merge upstream branch tips to obtain a shared upstream
                # parent.  This will add revision K (see graph above) to a
                # temporary merge target upstream tree.
                with tmp_target_utree.lock_write():
                    if us_ver > ut_ver:
                        # The source upstream tree is more recent and the
                        # temporary target tree needs to be reshaped to match
                        # it.
                        tmp_target_utree.revert(
                            None, source.repository.revision_tree(us_revid))
                        t_upstream_reverted = True

                    tmp_target_utree.set_parent_ids((ut_revid, us_revid))
                    new_revid = tmp_target_utree.commit(
                        'Prepared upstream tree for merging into target '
                        'branch.')

                    # Repository updates during a held lock are not visible,
                    # hence the call to refresh the data in the /target/ repo.
                    tree.branch.repository.refresh_data()

                    tree.branch.fetch(source, us_revid)
                    tree.branch.fetch(tmp_target_utree.branch, new_revid)

                    # Merge shared upstream parent into the target merge
                    # branch. This creates revison L in the digram above.
                    conflicts = tree.merge_from_branch(tmp_target_utree.branch)
                    if conflicts:
                        cmd = "bzr merge"
                        raise SharedUpstreamConflictsWithTargetPackaging(cmd)
                    else:
                        tree.commit(
                            'Merging shared upstream rev into target branch.')

    return (upstreams_diverged, t_upstream_reverted)


def report_fatal(code, description, *, hint=None):
    if os.environ.get('SVP_API') == '1':
        with open(os.environ['SVP_RESULT'], 'w') as f:
            json.dump({
                'result_code': code,
                'description': description}, f)
    logging.fatal('%s', description)
    if hint:
        logging.info('%s', hint)


def find_origins(source):
    for field, value in source.items():
        m = re.match(r'XS\-(.*)\-Vcs\-(.*)', field, re.I)
        if not m:
            continue
        vcs_type = m.group(2)
        if vcs_type == 'Browser':
            continue
        vcs_url = dict(
            vcs_field_to_bzr_url_converters)[vcs_type](value)
        origin = m.group(1)
        yield origin, vcs_type, vcs_url


def update_changelog(wt, subpath, target_distribution, version_fn,
                     summary, author=None):
    changes = []
    # TODO(jelmer): Iterate Build-Depends and verify that depends are
    # satisfied by target_distribution
    # TODO(jelmer): Update Vcs-Git/Vcs-Browser header?
    clp = wt.abspath(os.path.join(subpath, "debian/changelog"))

    if author is None:
        author = "%s <%s>" % get_maintainer()

    try:
        with ChangelogEditor(clp) as cl:
            # TODO(jelmer): If there was an existing backport, use that version
            cl.new_block(
                package=cl[0].package,
                distributions=target_distribution,
                urgency="low",
                author=author,
                date=format_date(),
                version=version_fn(cl[0].version),
                changes=[''],
            )
            changeblock_add_line(
                cl[0],
                [summary] +
                [" +" + line for line in changes],
            )
    except FileNotFoundError:
        raise MissingChangelogError([clp])
    except GeneratedFile as e:
        raise ChangelogGeneratedFile(e.path, e.template_path, e.template_type)

    debcommit(wt, subpath=subpath)


def backport_suffix(release):
    from distro_info import DebianDistroInfo
    distro_info = DebianDistroInfo()
    if release in ('stable', 'oldstable'):
        release = distro_info.codename(release)
    version = distro_info.version(release)
    assert version is not None
    return "bpo%s" % version


def determine_distribution(release: str, backport=False) -> str:
    if backport:
        if release.endswith('-backports'):
            return release
        from distro_info import DebianDistroInfo
        distro_info = DebianDistroInfo()
        if release == "stable" or distro_info.codename("stable") == release:
            return f"{distro_info.codename('stable')}-backports"
        if (release == "oldstable" or
                distro_info.codename("oldstable") == release):
            return f"{distro_info.codename('oldstable')}-backports-sloppy"
        raise Exception(f"unable to determine target suite for {release}")
    return release


def create_bpo_version(orig_version, bpo_suffix):
    m = re.fullmatch(r"(.*)\~" + bpo_suffix + r"\+([0-9]+)", str(orig_version))
    if m:
        base = m.group(1)
        buildno = int(m.group(2)) + 1
    else:
        base = str(orig_version)
        buildno = 1
    return f"{base}~{bpo_suffix}+{buildno}"


def auto_backport(argv=None):
    if argv is None:
        argv = sys.argv[1:]
    return main(argv + ['--backport'])


def main(argv=None):
    DEFAULT_BUILDER = "sbuild --no-clean-source"
    import argparse
    import breezy.bzr  # noqa: F401
    import breezy.git  # noqa: F401
    from breezy.workingtree import WorkingTree
    from breezy.workspace import check_clean_tree
    from breezy.branch import Branch

    from .apt_repo import RemoteApt, LocalApt, NoAptSources
    from .directory import source_package_vcs_url
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--directory', '-d', type=str, help="Working directory")
    parser.add_argument('--apt-repository', type=str,
                        help='APT Repository to fetch from')
    parser.add_argument('--apt-repository-key', type=str,
                        help='Repository key to use for verification')
    parser.add_argument('--version', type=str,
                        help='Version to use')
    parser.add_argument("--vendor", type=str,
                        help="Name of vendor to merge from")
    parser.add_argument("--backport", action="store_true")
    parser.add_argument("--package", type=str)
    parser.add_argument("--target-release", type=str)
    parser.add_argument('--build', action='store_true')
    # TODO(jelmer): add --revision argument
    parser.add_argument(
        "--builder",
        type=str,
        help="Build command",
        default=(
            DEFAULT_BUILDER + " --source --source-only-changes "
            "--debbuildopt=-v${LAST_VERSION}"
        ),
    )

    parser.add_argument('vcs_url', type=str, nargs='?')

    args = parser.parse_args(argv)

    logging.basicConfig(format='%(message)s', level=logging.INFO)

    wt, subpath = WorkingTree.open_containing(args.directory)

    check_clean_tree(wt, subpath=subpath)

    vendor = args.vendor

    try:
        cl, _larstiq = find_changelog(wt, subpath=subpath)
    except MissingChangelogError:
        cl = None
        since_version = None
    else:
        since_version = cl[0].version
        if not args.target_release:
            args.target_release = cl[0].distributions

    if args.vcs_url:
        branch_url = args.vcs_url
        vcs_type = None
        package = None
    elif args.package:
        package = args.package
        branch_url = None
        vcs_type = None
    else:
        if cl is None:
            report_fatal(
                'missing-changelog',
                'Need either --package, vcs url or existing debian package')
            return 1
        package = cl.package
        branch_url = None
        vcs_type = None

    if branch_url is None:
        try:
            with wt.get_file(os.path.join(subpath, 'debian/control')) as f:
                source = Deb822(f)
                origins = list(find_origins(source))
                if not origins:
                    report_fatal(
                        'no-upstream-vcs-url',
                        'Source package not have any Xs-*-Vcs-* fields')
                    return 1
                if len(origins) > 1:
                    logging.warning('More than one origin: %r', origins)
                    origin, vcs_type, branch_url = origins[0]
                else:
                    [(vendor, vcs_type, branch_url)] = origins
        except NoSuchFile:
            if args.apt_repository is not None:
                apt = RemoteApt.from_string(
                    args.apt_repository, args.apt_repository_key)
                origin = apt.distribution
            else:
                apt = LocalApt()
                origin = None

            logging.info('Using apt repository %r', apt)

            with apt:
                versions = []
                try:
                    for source in apt.iter_source_by_name(package):
                        versions.append((source['Version'], source))
                except NoAptSources:
                    report_fatal(
                        "no-apt-sources",
                        "No sources configured in /etc/apt/sources.list")
                    return 1

            versions.sort()
            try:
                version, source = versions[-1]
            except IndexError:
                report_fatal(
                    'not-present-in-apt',
                    f'The APT repository {apt} does not contain {package}')
                return 1

            source = next(apt.iter_source_by_name(package))
            vcs_type, branch_url = source_package_vcs_url(source)
            logging.info('Resolved apt repository to %s', branch_url)
            if not args.version:
                args.version = source['Version']

    assert branch_url

    branch = Branch.open(branch_url)

    if args.version is None:
        to_merge = branch.last_revision()
        wt.branch.repository.fetch(
            branch.repository, revision_id=to_merge)
        revtree = wt.branch.repository.revision_tree(to_merge)
        with revtree.get_file(os.path.join(subpath, 'debian/changelog')) as f:
            cl = Changelog(f, max_blocks=1)
        package = cl.package
        version = cl.version
    else:
        to_merge = None

    version = Version(args.version)

    if cl and version == cl.version:
        report_fatal(
            'tree-version-is-newer',
            f'Local tree already contains remote version {cl.version}')
        return 1

    if cl and version < cl.version:
        report_fatal(
            'nothing-to-do',
            f'Local tree contains newer version ({cl.version}) '
            f'than apt repo ({version})')
        return 1

    if cl and cl.version is not None:
        logging.info(
            'Importing version: %s (current: %s)', version, cl.version)
    else:
        logging.info('Importing version: %s', version)

    if to_merge is None:
        remote_db = DistributionBranch(branch, None)
        # Find the appropriate tag
        for tag_name in remote_db.possible_tags(version, vendor=vendor):
            try:
                to_merge = remote_db.branch.tags.lookup_tag(tag_name)
            except NoSuchTag:
                pass
            else:
                break
        else:
            report_fatal(
                'missing-remote-tag',
                'Unable to find tag for version {} in branch {}'.format(
                    version, remote_db.branch))
            return 1
        logging.info('Merging tag %s', tag_name)

    with wt.lock_write():
        try:
            wt.pull(branch, stop_revision=to_merge)
        except DivergedBranches:
            try:
                wt.merge_from_branch(branch, to_revision=to_merge)
            except ConflictsInTree as e:
                report_fatal('merged-conflicted', str(e))
                return 1
            except PointlessMerge as e:
                report_fatal('nothing-to-do', str(e))
                return 1
            except UnrelatedBranches:
                if apt:
                    logging.info(
                        'Upstream branch %r does not share history '
                        'with this one.', branch)
                    logging.info('Falling back to importing dsc.')
                    with tempfile.TemporaryDirectory() as td:
                        apt.retrieve_source(
                            cl.package, td, source_version=cl.version,
                            tar_only=False)
                        for entry in os.scandir(td):
                            if entry.name.endswith('.dsc'):
                                dsc_path = entry.path
                                break
                        else:
                            raise AssertionError(
                                f'{apt} did not actually '
                                'download dsc file') from None
                        local_db = DistributionBranch(wt.branch, None)
                        tag_name = local_db.import_package(dsc_path)
                        to_merge = wt.branch.tags.lookup_tag(tag_name)
                        try:
                            wt.merge_from_branch(
                                wt.branch, to_revision=to_merge)
                        except ConflictsInTree as e:
                            report_fatal('merged-conflicted', str(e))
                            return 1
                else:
                    report_fatal(
                        'unrelated-branches',
                        f'Upstream branch {branch} does not share '
                        'history with this one, and no apt repository '
                        'specified.')
                    return 1

        target_distribution = determine_distribution(
                args.target_release, args.backport)
        if args.backport:
            version_suffix = backport_suffix(args.target_release)

            def version_fn(imported_version):
                return create_bpo_version(imported_version, version_suffix)
            summary = f"Backport to {args.target_release}."
        else:
            if origin is not None:
                summary = f'Merge from {origin}.'
            else:
                origin = f'Merge {version}'

            version_fn = increment_version

        logging.info(
            "Using target distribution %s, version suffix %s",
            target_distribution,
            version_suffix,
        )
        update_changelog(wt, subpath, target_distribution=target_distribution,
                         version_fn=version_fn, summary=summary)

    if args.build:
        with tempfile.TemporaryDirectory() as td:
            builder = args.builder.replace(
                "${LAST_VERSION}", str(since_version))
            target_changes = _build_helper(
                wt, subpath, wt.branch, td, builder=builder
            )
            debsign(target_changes['source'])

            if not args.dry_run:
                dput_changes(target_changes['source'])

    if os.environ.get('SVP_API') == '1':
        with open(os.environ['SVP_RESULT'], 'w') as f:
            json.dump({
                'description': f"Merged from {origin}",
                'value': 80,
                'commit-message': f"Sync with {origin}",
                'context': {
                    'vendor': vendor,
                    'origin': origin,
                    'vcs_type': vcs_type,
                    'branch_url': branch_url,
                    'version': str(version),
                    'revision_id': to_merge.decode('utf-8'),
                    'tag': tag_name,
                },
            }, f)


if __name__ == '__main__':
    sys.exit(main(sys.argv[1:]))



#!/usr/bin/python3
# Copyright (C) 2018 Jelmer Vernooij <jelmer@jelmer.uk>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Refresh packaging."""

from email.utils import parseaddr
import logging
import os
from typing import Optional

from debian.changelog import get_maintainer
from debmutate.changelog import ChangelogEditor
from breezy.plugins.debian.changelog import debcommit
from breezy.tree import Tree


def override_dh_autoreconf_add_arguments(basedir: str, args):
    from debmutate._rules import update_rules

    # TODO(jelmer): Make sure dh-autoreconf is installed,
    # or debhelper version is >= 10

    def update_makefile(mf):
        for rule in mf.iter_rules(b"override_dh_autoreconf"):
            command = rule.commands()[0].split(b" ")
            if command[0] != b"dh_autoreconf":
                return
            rule.lines = [rule.lines[0]]
            command += args
            break
        else:
            rule = mf.add_rule(b"override_dh_autoreconf")
            command = [b"dh_autoreconf"] + args
        rule.append_command(b" ".join(command))

    return update_rules(
        makefile_cb=update_makefile,
        path=os.path.join(basedir, "debian", "rules")
    )


def update_packaging(
        tree: Tree, old_tree: Tree, subpath: str = "",
        committer: Optional[str] = None) -> list[str]:
    """Update packaging to take in changes between upstream trees.

    Args:
      tree: Current tree
      old_tree: Old tree
      committer: Optional committer to use for changes
    """
    if committer is None:
        maintainer = get_maintainer()
    else:
        maintainer = parseaddr(committer)
    notes = []
    tree_delta = tree.changes_from(old_tree, specific_files=[subpath])
    for delta in tree_delta.added:
        path = delta.path[1]
        if path is None:
            continue
        if not path.startswith(subpath):
            continue
        path = path[len(subpath):]
        if path == "autogen.sh":
            if override_dh_autoreconf_add_arguments(
                    tree.basedir, [b"./autogen.sh"]):
                logging.info(
                    "Modifying debian/rules: "
                    "Invoke autogen.sh from dh_autoreconf."
                )
                with ChangelogEditor(
                        tree.abspath(
                            os.path.join(subpath, 'debian/changelog'))) as cl:
                    cl.add_entry(
                        ["Invoke autogen.sh from dh_autoreconf."],
                        maintainer=maintainer)
                debcommit(
                    tree,
                    committer=committer,
                    subpath=subpath,
                    paths=["debian/changelog", "debian/rules"],
                )
        elif path.startswith("LICENSE") or path.startswith("COPYING"):
            notes.append(
                "License file %s has changed." % os.path.join(subpath, path))

    return notes


def main():
    import argparse
    from breezy.workingtree import WorkingTree
    from breezy.revisionspec import RevisionSpec

    import breezy.bzr  # noqa: F401
    import breezy.git  # noqa: F401

    parser = argparse.ArgumentParser('deb-update-packaging')
    parser.add_argument(
        '--since', type=str, help='Revision since when to update')
    args = parser.parse_args()

    logging.basicConfig(level=logging.INFO, format='%(message)s')

    wt, subpath = WorkingTree.open_containing('.')
    if args.since:
        old_tree = RevisionSpec.from_string(args.since).as_tree(args.since)
    else:
        old_tree = wt.basis_tree()

    notes = update_packaging(wt, old_tree, subpath)
    for note in notes:
        logging.info('%s', note)


if __name__ == "__main__":
    import sys

    sys.exit(main())



#!/usr/bin/python3
# Copyright (C) 2018 Jelmer Vernooij <jelmer@jelmer.uk>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import contextlib
import errno
import json
from hashlib import sha1
import logging
import os
import subprocess
import tempfile
from typing import Optional

from urllib.request import urlopen
from urllib.error import HTTPError, URLError

from debmutate.control import ControlEditor
from debmutate.changelog import ChangelogEditor, distribution_is_unreleased

import breezy.bzr  # noqa: F401
import breezy.git  # noqa: F401
from breezy import urlutils
from breezy.errors import (
    NotBranchError, NoSuchTag, ConflictsInTree, NoSuchRevisionInTree)
from breezy.revision import RevisionID, NULL_REVISION
from breezy.trace import note, warning
from breezy.transform import MalformedTransform
from breezy.tree import Tree

from breezy.transport import NoSuchFile
from breezy.workingtree import WorkingTree

from debian.changelog import Changelog, Version

from .changelog import debcommit
from .directory import vcs_git_url_to_bzr_url
from .info import versions_dict
from .upstream import PackageVersionNotPresent
from .import_dsc import (
    DistributionBranch,
    DistributionBranchSet,
    VersionAlreadyImported,
)
from .apt_repo import (
    Apt,
    LocalApt,
    RemoteApt,
    NoAptSources,
    AptSourceError,
)

BRANCH_NAME = "missing-commits"


def connect_udd_mirror():
    import psycopg2

    return psycopg2.connect(
        database="udd",
        user="udd-mirror",
        password="udd-mirror",
        host="udd-mirror.debian.net",
    )


def select_vcswatch_packages():
    conn = connect_udd_mirror()
    cursor = conn.cursor()
    args = []
    query = """\
    SELECT sources.source, vcswatch.url
    FROM vcswatch JOIN sources ON sources.source = vcswatch.source
    WHERE
     vcswatch.status IN ('OLD', 'UNREL') AND
     sources.release = 'sid'
"""
    cursor.execute(query, tuple(args))
    packages = []
    for package, vcs_url in cursor.fetchall():
        packages.append(package)
    return packages


class SnapshotDownloadError(Exception):

    def __init__(self, url, inner, transient):
        self.url = url
        self.inner = inner
        self.transient = transient


class SnapshotMissing(Exception):

    def __init__(self, source_name, source_version):
        self.source_name = source_name
        self.source_version = source_version


class SnapshotHashMismatch(Exception):

    def __init__(self, filename, actual_hash, expected_hash):
        self.filename = filename
        self.actual_hash = actual_hash
        self.expected_hash = expected_hash


def download_snapshot(package: str, version: Version, output_dir: str) -> str:
    note("Downloading %s %s", package, version)
    srcfiles_url = (
        "https://snapshot.debian.org/mr/package/%s/%s/"
        "srcfiles?fileinfo=1" % (package, version)
    )
    files = {}
    try:
        srcfiles = json.load(urlopen(srcfiles_url))
    except HTTPError as e:
        if e.status == 404:
            raise SnapshotMissing(package, version) from e
        if e.status // 100 == 5:
            transient = True
        else:
            transient = None
        raise SnapshotDownloadError(
            srcfiles_url, e, transient=transient) from e
    except URLError as e:
        if e.errno == errno.ENETUNREACH:
            transient = True
        else:
            transient = None
        raise SnapshotDownloadError(
            srcfiles_url, e, transient=transient) from e

    for hsh, entries in srcfiles["fileinfo"].items():
        for entry in entries:
            files[entry["name"]] = hsh
    for filename, hsh in files.items():
        local_path = os.path.join(output_dir, os.path.basename(filename))
        try:
            with open(local_path, "rb") as f:
                actual_hsh = sha1(f.read()).hexdigest()
            if actual_hsh != hsh:
                raise SnapshotHashMismatch(filename, actual_hsh, hsh)
        except FileNotFoundError:
            with open(local_path, "wb") as f:
                url = "https://snapshot.debian.org/file/%s" % hsh
                note('.. Downloading %s -> %s', url, filename)
                try:
                    with urlopen(url) as g:
                        f.write(g.read())
                except HTTPError as e:
                    if e.status // 100 == 5:
                        transient = True
                    else:
                        transient = None
                    raise SnapshotDownloadError(
                        url, e, transient=transient) from e
                except URLError as e:
                    if e.errno == errno.ENETUNREACH:
                        transient = True
                    else:
                        transient = None
                    raise SnapshotDownloadError(
                        url, e, transient=None) from e
    file_version = Version(version)
    file_version.epoch = None
    dsc_filename = "{}_{}.dsc".format(package, file_version)
    return os.path.join(output_dir, dsc_filename)


class NoopChangesOnly(Exception):
    def __init__(self, vcs_version, archive_version):
        self.vcs_version = vcs_version
        self.archive_version = archive_version
        super(NoopChangesOnly, self).__init__(
            "No missing versions with effective changes. "
            "Archive has %s, VCS has %s"
            % (archive_version, vcs_version)
        )


class NoMissingVersions(Exception):
    def __init__(self, vcs_version, archive_version):
        self.vcs_version = vcs_version
        self.archive_version = archive_version
        super().__init__(
            "No missing versions after all. Archive has %s, VCS has %s"
            % (archive_version, vcs_version)
        )


class TreeVersionNotInArchiveChangelog(Exception):
    def __init__(self, tree_version):
        self.tree_version = tree_version
        super().__init__(
            "tree version %s does not appear in archive changelog" %
            tree_version
        )


class TreeVersionWithoutTag(Exception):
    def __init__(self, tree_version, tag_name):
        self.tree_version = tree_version
        super().__init__(
            "unable to find revision for version {}; no tags (e.g. {})".format(
                tree_version, tag_name))


class TreeUpstreamVersionMissing(Exception):
    def __init__(self, upstream_version):
        self.upstream_version = upstream_version
        super().__init__(
            "unable to find upstream version %r" % upstream_version
        )


class UnreleasedChangesSinceTreeVersion(Exception):
    def __init__(self, tree_version):
        super().__init__(
            "there are unreleased changes since %s" % tree_version)


def find_missing_versions(
        archive_cl: Version, tree_version: Version) -> list[Version]:
    missing_versions: list[Version] = []
    for block in archive_cl:
        if tree_version is not None and block.version == tree_version:
            break
        missing_versions.append(block.version)
    else:
        if tree_version is not None:
            raise TreeVersionNotInArchiveChangelog(tree_version)
    return missing_versions


def is_noop_upload(tree, basis_tree=None, subpath=''):
    if basis_tree is None:
        basis_tree = tree.basis_tree()
    changes = tree.iter_changes(basis_tree)
    try:
        while True:
            change = next(changes)
            if change.path[1] != "":
                break
    except StopIteration:
        return True
    cl_path = os.path.join(subpath, 'debian', 'changelog')
    if change.path != (cl_path, cl_path):
        return False
    # if there are any other changes, then this is not trivial:
    try:
        next(changes)
    except StopIteration:
        pass
    else:
        return False
    try:
        new_cl = Changelog(tree.get_file_text(cl_path))
    except NoSuchFile:
        return False
    try:
        old_cl = Changelog(basis_tree.get_file_text(cl_path))
    except NoSuchFile:
        return False

    del new_cl._blocks[0]
    # TODO(jelmer): Check for uploads that aren't just meant to trigger a
    # build.  i.e. closing bugs.
    return str(new_cl) == str(old_cl)


def import_uncommitted(
        tree: Tree, subpath: str, apt: Apt,
        source_name: str,
        archive_version: Optional[Version] = None,
        tree_version: Optional[Version] = None,
        merge_unreleased: bool = True,
        skip_noop: bool = True,
        ) -> list[tuple[str, Version, RevisionID]]:
    with contextlib.ExitStack() as es:
        es.enter_context(apt)
        archive_source = es.enter_context(tempfile.TemporaryDirectory())
        apt.retrieve_source(
            source_name, archive_source, source_version=archive_version)
        [dsc] = [e.name for e in os.scandir(archive_source)
                 if e.name.endswith('.dsc')]
        note("Unpacking source %s", dsc)
        subprocess.check_output(['dpkg-source', '-x', dsc], cwd=archive_source)
        [subdir] = [e.path for e in os.scandir(archive_source) if e.is_dir()]
        with open(os.path.join(subdir, "debian", "changelog")) as f:
            archive_cl = Changelog(f)
        missing_versions = find_missing_versions(archive_cl, tree_version)
        if len(missing_versions) == 0:
            raise NoMissingVersions(tree_version, archive_cl.version)
        note("Missing versions: %s", ", ".join(map(str, missing_versions)))

        ret = []
        dbs = DistributionBranchSet()
        db = DistributionBranch(tree.branch, tree.branch, tree=tree)
        dbs.add_branch(db)

        if tree_version is not None:
            try:
                tree_version_revid = db.revid_of_version(tree_version)
            except NoSuchTag as e:
                raise TreeVersionWithoutTag(tree_version, e.tag_name) from e
            if tree_version_revid != tree.last_revision():
                # There are changes since the last tree version.
                note(
                    "Commits exist on the branch since last upload to archive")
                if not merge_unreleased:
                    raise UnreleasedChangesSinceTreeVersion(tree_version)

                merge_into = tree.last_revision()
                tree.update(revision=tree_version_revid)
            else:
                merge_into = None
        else:
            merge_into = None

        applied_patches = tree.has_filename(".pc/applied-patches")
        if tree_version and tree_version.debian_revision:
            try:
                upstream_tips = (
                    db.pristine_upstream_source.version_as_revisions(
                        source_name, tree_version.upstream_version))
            except PackageVersionNotPresent as e:
                raise TreeUpstreamVersionMissing(
                    tree_version.upstream_version) from e
            else:
                note("Extracting upstream version %s.",
                     tree_version.upstream_version)
                upstream_dir = es.enter_context(tempfile.TemporaryDirectory())
                db.extract_upstream_tree(upstream_tips, upstream_dir)
        else:
            upstream_dir = es.enter_context(tempfile.TemporaryDirectory())
            db.create_empty_upstream_tree(upstream_dir)
        output_dir = es.enter_context(tempfile.TemporaryDirectory())
        last_revid = db.tree.last_revision()
        for version in reversed(missing_versions):
            try:
                dsc_path = download_snapshot(source_name, version, output_dir)
            except SnapshotMissing as e:
                warning('Missing snapshot for %s (never uploaded?), skipping.',
                        e.source_version)
                continue
            note("Importing %s", version)
            try:
                tag_name = db.import_package(
                    dsc_path, apply_patches=applied_patches)
            except VersionAlreadyImported as e:
                # Present in the repository, just not on the branch
                note("%s was already imported (tag: %s), just not on the "
                     "branch. Updating tree.", e.version, e.tag_name)
                tag_name = e.tag_name
                db.tree.update(revision=db.branch.tags.lookup_tag(e.tag_name))
            revid = db.branch.tags.lookup_tag(tag_name)
            if skip_noop and last_revid != NULL_REVISION:
                try:
                    last_tree = db.tree.revision_tree(last_revid)
                except NoSuchRevisionInTree:
                    last_tree = db.branch.repository.revision_tree(last_revid)
                if is_noop_upload(tree, last_tree, subpath):
                    note('Skipping version %s without effective changes',
                         version)
                    tree.update(revision=last_revid)
                    continue
            ret.append((tag_name, version, revid))
            last_revid = revid

    if not ret:
        raise NoopChangesOnly(tree_version, archive_cl.version)

    if merge_into:
        to_merge = tree.last_revision()
        tree.update(revision=merge_into)
        tree.merge_from_branch(tree.branch, to_revision=to_merge)
        revid = debcommit(
            tree, subpath=subpath,
            message='Merge archive versions: %s' % ', '.join([
                str(v) for (t, v, r) in ret]))
        parent_ids = tree.branch.repository.get_revision(revid).parent_ids
        assert parent_ids == [merge_into, to_merge], \
            "Expected parents to be {!r}, was {!r}".format(
                [merge_into, to_merge], parent_ids)
    return ret


def report_fatal(code, description, *, hint=None, transient=None):
    if os.environ.get('SVP_API') == '1':
        with open(os.environ['SVP_RESULT'], 'w') as f:
            json.dump({
                'versions': versions_dict(),
                'transient': transient,
                'result_code': code,
                'description': description}, f)
    logging.fatal('%s', description)
    if hint:
        logging.info('%s', hint)


def set_vcs_git_url(control, vcs_git_base: Optional[str],
                    vcs_browser_base: Optional[str]):
    old_vcs_url = control.source.get("Vcs-Git")
    if vcs_git_base is not None:
        control.source["Vcs-Git"] = urlutils.join(
            vcs_git_base, "%s.git" % control.source['Source'])
    new_vcs_url = control.source.get("Vcs-Git")
    if vcs_browser_base:
        control.source["Vcs-Browser"] = urlutils.join(
            vcs_browser_base, control.source['Source'])
    return (old_vcs_url, new_vcs_url)


def contains_git_attributes(tree, subpath):
    for path, versioned, kind, ie in tree.list_files(
            recursive=True, recurse_nested=True, from_dir=subpath):
        if os.path.basename(path) == '.gitattributes':
            return True
    return False


def main(argv=None):
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--apt-repository', type=str,
        help='APT repository to use. Defaults to locally configured.',
        default=(
            os.environ.get('APT_REPOSITORY')
            or os.environ.get('REPOSITORIES')))
    parser.add_argument(
        '--apt-repository-key', type=str,
        help=('APT repository key to use for validation, '
              'if --apt-repository is set.'),
        default=os.environ.get('APT_REPOSITORY_KEY'))
    parser.add_argument(
        '--version', type=str,
        help='Source version to import')
    parser.add_argument(
        '--vcs-git-base', type=str,
        help='Set Vcs-Git URL')
    parser.add_argument(
        '--vcs-browser-base', type=str,
        help='Set Vcs-Browser URL')
    parser.add_argument(
        '--no-merge-unreleased',
        action='store_true',
        help=('Error rather than merge when there are '
              'unreleased changes'))
    parser.add_argument(
        '--no-skip-noop',
        action='store_true',
        help='Do not skip uploads without effective changes')
    parser.add_argument(
        '--package', type=str, help='Package to import',
        default=os.environ.get('PACKAGE'))
    parser.add_argument(
        '--force-git-attributes', action='store_true',
        help='Force importing even if the tree contains git attributes')
    args = parser.parse_args(argv)

    logging.basicConfig(level=logging.INFO, format='%(message)s')

    if args.apt_repository:
        apt = RemoteApt.from_string(
            args.apt_repository, args.apt_repository_key)
    else:
        apt = LocalApt()
    try:
        local_tree, subpath = WorkingTree.open_containing('.')
    except NotBranchError:
        report_fatal(
            "not-branch-error",
            "Not running in a version-controlled directory")
        return 1

    cl_path = os.path.join(subpath, "debian/changelog")
    try:
        with local_tree.get_file(cl_path) as f:
            tree_cl = Changelog(f)
            source_name = tree_cl.package
            for block in tree_cl:
                if distribution_is_unreleased(block.distributions):
                    continue
                tree_version = block.version
                break
            else:
                tree_version = None
    except NoSuchFile as e:
        if local_tree.last_revision() == NULL_REVISION and args.package:
            source_name = args.package
            tree_version = None
            tree_cl = None
        else:
            if local_tree.last_revision() == NULL_REVISION:
                hint = "Tree is empty. Specify --package?"
            else:
                hint = None
            report_fatal(
                "missing-changelog",
                "Missing changelog: %s" % e.path,
                hint=hint)
            return 1
    else:
        if args.package and tree_cl.package != args.package:
            report_fatal(
                'inconsistent-package',
                'Inconsistent package name: {} specified, {} found'.format(
                    args.package, tree_cl.package))
            return 1

    if not args.force_git_attributes and hasattr(
            local_tree.branch.repository, '_git'):
        # See https://salsa.debian.org/jelmer/janitor.debian.net/-/issues/74
        if contains_git_attributes(local_tree, subpath):
            report_fatal(
                'unsupported-git-attributes',
                'Tree contains .gitattributes which may impact imports and '
                'are unsupported',
                hint='Run with --force-git-attributes to ignore')
            return 1

    try:
        ret = import_uncommitted(
            local_tree, subpath, apt, source_name=source_name,
            archive_version=Version(args.version) if args.version else None,
            tree_version=tree_version,
            merge_unreleased=not args.no_merge_unreleased,
            skip_noop=not args.no_skip_noop)
    except AptSourceError as e:
        if isinstance(e.reason, list):
            reason = e.reason[-1]
        else:
            reason = e.reason
        report_fatal("apt-source-error", reason)
        return 1
    except NoAptSources:
        report_fatal(
            "no-apt-sources",
            "No sources configured in /etc/apt/sources.list")
        return 1
    except TreeVersionWithoutTag as e:
        report_fatal("tree-version-not-found", str(e))
        return 1
    except TreeUpstreamVersionMissing as e:
        report_fatal("tree-upstream-version-missing", str(e))
        return 1
    except UnreleasedChangesSinceTreeVersion as e:
        report_fatal("unreleased-changes", str(e))
        return 1
    except TreeVersionNotInArchiveChangelog as e:
        report_fatal("tree-version-not-in-archive-changelog", str(e))
        return 1
    except NoopChangesOnly as e:
        report_fatal(
            "nothing-to-do", str(e),
            hint='Run with --no-skip-noop to include trivial uploads.')
        return 1
    except NoMissingVersions as e:
        report_fatal("nothing-to-do", str(e))
        return 1
    except SnapshotDownloadError as e:
        report_fatal(
            'snapshot-download-failed',
            'Downloading {} failed: {}'.format(e.url, e.inner),
            transient=e.transient)
        return 1
    except SnapshotHashMismatch as e:
        report_fatal(
            'snapshot-hash-mismatch',
            'Snapshot hash mismatch for {}: {} != {}'.format(
                e.filename, e.expected_hash, e.actual_hash))
        return 1
    except MalformedTransform as e:
        report_fatal('malformed-transform', str(e))
        return 1
    except ConflictsInTree:
        report_fatal(
            'merge-conflicts',
            "Merging uncommitted changes resulted in conflicts.",
            transient=False)
        return 1

    if args.vcs_git_base:
        with ControlEditor(local_tree.abspath('debian/control')) as control:
            (old_vcs_url, new_vcs_url) = set_vcs_git_url(
                 control, args.vcs_git_base, args.vcs_browser_base)
        if old_vcs_url != new_vcs_url:
            note('Updating Vcs-Git URL to %s', new_vcs_url)
            with ChangelogEditor(
                    local_tree.abspath('debian/changelog')) as changelog:
                changelog.add_entry(['Set Vcs-Git header.'])
            debcommit(local_tree, subpath=subpath)
            target_branch_url = vcs_git_url_to_bzr_url(new_vcs_url)
        else:
            target_branch_url = None
    else:
        target_branch_url = None

    if os.environ.get('SVP_API') == '1':
        if len(ret) == 1:
            commit_message = "Import missing upload: %s" % ret[0][1]
            description = 'Import uploaded version: %s' % (
                ret[0][1])
        else:
            commit_message = "Import missing uploads: %s." % (
                    ", ".join([str(v) for t, v, rs in ret]))
            description = 'Import uploaded versions: %r' % (
                [str(v) for t, v, rs in ret])
        with open(os.environ['SVP_RESULT'], 'w') as f:
            json.dump({
                'description': description,
                'versions': versions_dict(),
                'value': 60 + sum(
                     [60 if 'nmu' in str(e[1]) else 20 for e in ret]),
                'commit-message': commit_message,
                'context': {
                    'versions': [tag_name for (tag_name, version, rs) in ret],
                    'tags':
                        [(tag_name, str(version))
                         for (tag_name, version, rs) in ret],
                },
                'target-branch-url': target_branch_url,
            }, f)

    note('Imported uploads: %s.', [str(v[1]) for v in ret])


if __name__ == "__main__":
    import sys

    sys.exit(main(sys.argv[1:]))



#    launchpad.py -- Lookups via. launchpadlib
#    Copyright (C) 2009 Canonical Ltd.
#
#    This file is part of bzr-builddeb.
#
#    bzr-builddeb is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    bzr-builddeb is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with bzr-builddeb; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#

from ...trace import (
    mutter,
)

try:
    from launchpadlib.launchpad import Launchpad
    from launchpadlib.uris import LPNET_SERVICE_ROOT
    HAVE_LPLIB = True
except ImportError:
    HAVE_LPLIB = False


def get_launchpad():
    if not HAVE_LPLIB:
        return None
    return Launchpad.login_anonymously(
        "breezy-debian", service_root=LPNET_SERVICE_ROOT)


def ubuntu_bugs_for_debian_bug(bug_id):
    """Find Ubuntu bugs linked to a particular Debian bug.

    :param bug_id: Debian bug id
    :return: Liust of Launchpad bug ids for Ubuntu bugs
    """
    lp = get_launchpad()
    if lp is None:
        return []
    try:
        bug = lp.load(str(lp._root_uri) + "bugs/bugtrackers/debbugs/%s")
        tasks = bug.bug_tasks
        for task in tasks:
            if task.bug_target_name.endswith("(Ubuntu)"):
                return str(bug.id)
    except Exception as e:
        mutter(str(e))
        return []
    return []


def debian_bugs_for_ubuntu_bug(bug_id):
    """Find the Debian bugs linked to a particular Ubuntu bug.

    :param bug_id: The Launchpad bug ID for the Ubuntu bug
    :return: List of Debian bug URLs.
    """
    lp = get_launchpad()
    if lp is None:
        return []
    try:
        bug = lp.bugs[int(bug_id)]
        tasks = bug.bug_tasks
        for task in tasks:
            if task.bug_target_name.endswith("(Debian)"):
                watch = task.bug_watch
                if watch is None:
                    break
                return watch.remote_bug.encode("utf-8")
    except Exception as e:
        mutter(str(e))
        return []
    return []



#    test_builddeb.py -- Blackbox tests for builddeb.
#    Copyright (C) 2007 James Westby <jw+debian@jameswestby.net>
#
#    This file is part of bzr-builddeb.
#
#    bzr-builddeb is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    bzr-builddeb is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with bzr-builddeb; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#

import os

from .. import BuilddebTestCase


class TestBuilddeb(BuilddebTestCase):

    commited_file = 'commited_file'
    uncommited_file = 'uncommited_file'
    unadded_file = 'unadded_file'

    def make_unpacked_source(self):
        """Create an unpacked source tree in a branch.

        Return the working tree"""
        tree = self.make_branch_and_tree('.')
        cl_file = 'debian/changelog'
        source_files = ['README', 'debian/'] + [cl_file]
        self.build_tree(source_files)
        c = self.make_changelog()
        self.write_changelog(c, cl_file)
        tree.add(source_files)
        return tree

    def build_dir(self):
        return os.path.join('..', 'build-area',
                            self.package_name + '-' +
                            str(self.upstream_version))

    def assertInBuildDir(self, files):
        build_dir = self.build_dir()
        if not isinstance(files, list):
            files = [files]
        for filename in files:
            self.assertPathExists(os.path.join(build_dir, filename))

    def assertNotInBuildDir(self, files):
        build_dir = self.build_dir()
        if not isinstance(files, list):
            files = [files]
        for filename in files:
            self.assertPathDoesNotExist(os.path.join(build_dir, filename))

    def test_builddeb_registered(self):
        self.run_bzr("builddeb --help")

    def test_bd_alias(self):
        self.run_bzr("bd --help")

    def test_builddeb_not_package(self):
        self.run_bzr_error(['Could not find changelog'], 'builddeb')

    def build_really_simple_tree(self):
        tree = self.make_unpacked_source()
        self.build_tree([self.commited_file, self.uncommited_file,
                         self.unadded_file])
        tree.add([self.commited_file])
        tree.commit("one", rev_id=b'revid1')
        tree.add([self.uncommited_file])
        return tree

    def build_tree_with_conflict(self):
        tree = self.make_unpacked_source()
        self.build_tree([self.commited_file, self.uncommited_file,
                         self.unadded_file])
        tree.add([self.commited_file])
        tree.commit("one", rev_id=b'revid1')
        newtree = tree.controldir.sprout('newtree').open_workingtree()
        tree.add([self.uncommited_file])
        tree.commit("two", rev_id=b'revid2')

        p = '{}/work/newtree/{}'.format(
            self.test_base_dir, self.uncommited_file)
        with open(p, 'w') as fh:
            fh.write('** This is the conflicting line.')
        newtree.add([self.uncommited_file])
        newtree.commit("new-two", rev_id=b'revidn2')
        conflicts = tree.merge_from_branch(newtree.branch)

        return (conflicts, tree)

    def test_builddeb_uses_working_tree(self):
        self.build_really_simple_tree()
        self.run_bzr("builddeb --native --builder true --dont-purge")
        self.assertInBuildDir([self.commited_file, self.uncommited_file])
        self.assertNotInBuildDir([self.unadded_file])

    def test_builddeb_refuses_tree_with_conflicts(self):
        (conflicts, tree) = self.build_tree_with_conflict()
        self.assertTrue(conflicts)
        self.run_bzr_error(
            ['There are conflicts in the working tree. '
             'You must resolve these'],
            "builddeb --native --builder true --dont-purge")

    def test_builddeb_strict(self):
        self.build_really_simple_tree()
        self.run_bzr_error(
            ["brz: ERROR: Build refused because there are unknown "
             "files in the tree. "
             "To list all known files, run 'bzr unknowns'."],
            "builddeb --strict")

    def test_builddeb_uses_revision_when_told(self):
        self.build_really_simple_tree()
        self.run_bzr("builddeb --native --builder true --dont-purge -r-1")
        self.assertInBuildDir([self.commited_file])
        self.assertNotInBuildDir([self.unadded_file, self.uncommited_file])

    def test_builddeb_error_on_two_revisions(self):
        self.make_unpacked_source()
        self.run_bzr_error(
            ['--revision takes exactly one revision specifier.'],
            "builddeb --native --builder true -r0..1")

    def test_builddeb_allows_building_revision_0(self):
        self.build_really_simple_tree()
        # This may break if there is something else that needs files in the
        # branch before the changelog is looked for.
        self.run_bzr_error(['Could not find changelog'],
                           "builddeb --native --builder true --dont-purge -r0")
        self.assertNotInBuildDir([self.commited_file, self.unadded_file,
                                  self.uncommited_file])

    def orig_dir(self):
        return '..'

    def make_upstream_tarball(self):
        f = open(os.path.join(self.orig_dir(), self.package_name + "_" +
                              str(self.package_version.upstream_version) +
                              ".orig.tar.gz"), 'wb')
        f.close()

    def test_builder(self):
        self.make_unpacked_source()
        self.run_bzr('bd --dont-purge --native --builder "touch built"')
        self.assertInBuildDir('built')

    def test_package_merge_first(self):
        self.make_unpacked_source()
        self.make_upstream_tarball()
        self.run_bzr('bd --package-merge --builder true')

    def test_hooks(self):
        self.make_unpacked_source()
        self.make_upstream_tarball()
        os.mkdir('.bzr-builddeb/')
        with open('.bzr-builddeb/default.conf', 'w') as f:
            f.write('[HOOKS]\npre-export = touch pre-export\n')
            f.write('pre-build = touch pre-build\n'
                    'post-build = touch post-build\n')
        self.run_bzr('add .bzr-builddeb/default.conf')
        self.run_bzr('bd --dont-purge --builder true')
        self.assertPathExists('pre-export')
        self.assertInBuildDir(['pre-build', 'post-build'])

    def test_utf8_changelog(self):
        from ... import debian_changelog_commit
        from .....msgeditor import hooks
        hooks.install_named_hook(
            "set_commit_message",
            debian_changelog_commit, "Test builddeb set commit msg hook")
        tree = self.make_unpacked_source()
        # The changelog is only used for commit message when it already exists
        # and is then changed, so need to clear it, commit, then set the
        # contents.
        with open("debian/changelog", "w") as f:
            pass
        tree.commit("Prepare for release")
        # Would be nice to use debian.changelog to make one, but it's changed
        # across versions as to how it wants non-ascii bytes provided.
        with open("debian/changelog", "wb") as f:
            f.write(b"  * \xe2\x80\xa6and another thing")
        self.run_bzr(['commit'])
        branch = tree.branch
        self.assertEqual(
            "\u2026and another thing",
            branch.repository.get_revision(branch.last_revision()).message)



#!/usr/bin/python3
# Copyright (C) 2018 Jelmer Vernooij <jelmer@jelmer.uk>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Support for merging new upstream versions."""

import errno
import json
import logging
import os
import socket
import ssl
import tempfile
import traceback
from typing import Optional, Callable, Union

from breezy.bzr import LineEndingError
from breezy.git.remote import RemoteGitError
import breezy.plugins.launchpad  # noqa: F401
from breezy.workingtree import WorkingTree

from debian.changelog import Version, ChangelogParseError

from breezy import errors, urlutils

from breezy.branch import Branch
from breezy.controldir import ControlDir
from breezy.errors import (
    InvalidNormalization,
    NoCommits,
    InvalidHttpResponse,
    NoRoundtrippingSupport,
    UncommittedChanges,
)
from breezy.transport import (
    FileExists,
    NoSuchFile,
    UnsupportedProtocol,
)

from breezy.workingtree import PointlessMerge
from breezy.transport import (
    Transport,
    UnusableRedirect,
    get_transport,
    )
from .config import UpstreamMetadataSyntaxError
from .info import versions_dict
from .util import (
    InconsistentSourceFormatError,
)
from .import_dsc import (
    UpstreamAlreadyImported,
    UpstreamBranchAlreadyMerged,
    CorruptUpstreamSourceFile,
)
from .changelog import debcommit

from breezy.transform import MalformedTransform

from .merge_upstream import (
    get_upstream_branch_location,
    do_import,
    do_merge,
    get_existing_imported_upstream_revids,
    get_tarballs,
)
from .repack_tarball import (
    UnsupportedRepackFormat,
)

from .upstream.pristinetar import (
    PristineTarError,
    get_pristine_tar_source,
)

from .util import (
    debuild_config,
    guess_build_type,
    get_files_excluded,
    tree_contains_upstream_source,
    BUILD_TYPE_MERGE,
    BUILD_TYPE_NATIVE,
    find_changelog,
    MissingChangelogError,
    control_files_in_root,
    full_branch_url,
)

from .upstream import (
    TarfileSource,
    MissingUpstreamTarball,
    PackageVersionNotPresent,
)
from .upstream.uscan import (
    UScanSource,
    UScanError,
    NoWatchFile,
    WatchLineWithoutMatches,
    WatchLineWithoutMatchingHrefs,
)
from .upstream.branch import (
    UpstreamBranchSource,
    DistCommandFailed,
    run_dist_command,
    PreviousVersionTagMissing,
)

from debmutate.changelog import ChangelogEditor, upstream_merge_changelog_line
from debmutate.reformatting import GeneratedFile
from debmutate.versions import (
    add_dfsg_suffix,
    strip_dfsg_suffix,
    matches_release,
    new_upstream_package_version,
    initial_debian_revision,
    debianize_upstream_version,
    )

from debmutate.vcs import split_vcs_url
from debmutate.watch import WatchSyntaxError

from breezy.tree import Tree, MissingNestedTree


class BigVersionJump(Exception):
    """There was a big version jump."""

    def __init__(self, old_upstream_version, new_upstream_version):
        self.old_upstream_version = old_upstream_version
        self.new_upstream_version = new_upstream_version


class DistMissingNestedTree(Exception):
    """Dist failed to find a nested tree."""

    def __init__(self, path):
        self.path = path


class UpstreamMergeConflicted(Exception):
    """The upstream merge resulted in conflicts."""

    def __init__(self, upstream_version, conflicts):
        self.version = upstream_version
        self.conflicts = conflicts


class UpstreamAlreadyMerged(Exception):
    """Upstream release (or later version) has already been merged."""

    def __init__(self, upstream_version):
        self.version = upstream_version


class UpstreamBranchLocationInvalid(Exception):
    """Upstream branch location is invalid."""

    def __init__(self, url, extra):
        self.url = url
        self.extra = extra


class InvalidFormatUpstreamVersion(Exception):
    """Invalid format upstream version string."""

    def __init__(self, version, source):
        self.version = version
        self.source = source


class UpstreamBranchUnknown(Exception):
    """The location of the upstream branch is unknown."""


class NewUpstreamMissing(Exception):
    """Unable to find upstream version to merge."""


class UpstreamBranchUnavailable(Exception):
    """Snapshot merging was requested by upstream branch is unavailable."""

    def __init__(self, location, error):
        self.location = location
        self.error = error


class UpstreamVersionMissingInUpstreamBranch(Exception):
    """The upstream version is missing in the upstream branch."""

    def __init__(self, upstream_branch, upstream_version):
        self.branch = upstream_branch
        self.version = upstream_version


class PackageIsNative(Exception):
    """Unable to merge upstream version."""

    def __init__(self, package, version):
        self.package = package
        self.version = version


class UpstreamNotBundled(Exception):
    """Packaging branch does not carry upstream sources."""

    def __init__(self, package):
        self.package = package


class NewUpstreamTarballMissing(Exception):
    def __init__(self, package, version, upstream):
        self.package = package
        self.version = version
        self.upstream = upstream


class NoUpstreamLocationsKnown(Exception):
    """No upstream locations (uscan/repository) for the package are known."""

    def __init__(self, package):
        self.package = package


class NewerUpstreamAlreadyImported(Exception):
    """A newer upstream version has already been imported."""

    def __init__(self, old_upstream_version, new_upstream_version):
        self.old_upstream_version = old_upstream_version
        self.new_upstream_version = new_upstream_version


class ChangelogGeneratedFile(Exception):
    """The changelog file is generated."""

    def __init__(self, path, template_path, template_type):
        self.path = path
        self.template_path = template_path
        self.template_type = template_type


class ReleaseWithoutChanges(Exception):
    """There was a new release, but it didn't change anything."""

    def __init__(self, new_upstream_version):
        self.new_upstream_version = new_upstream_version


class MergingUpstreamSubpathUnsupported(Exception):
    """Merging an upstream branch with subpath is not yet supported."""


RELEASE_BRANCH_NAME = "new-upstream-release"
SNAPSHOT_BRANCH_NAME = "new-upstream-snapshot"
DEFAULT_DISTRIBUTION = "unstable"
VALUE_IMPORT = {
    'release': 20,
    'snapshot': 10
}
VALUE_MERGE = {
    'release': 40,
    'snapshot': 30,
}


def is_big_version_jump(old_upstream_version, new_upstream_version):
    try:
        old_major_version = int(str(old_upstream_version).split('.')[0])
    except ValueError:
        return False
    try:
        new_major_version = int(str(new_upstream_version).split('.')[0])
    except ValueError:
        return False
    if old_major_version > 0 and new_major_version > 5*old_major_version:
        return True
    return False


class ImportUpstreamResult:
    """Object representing the result of an import_upstream operation."""

    __slots__ = [
        "old_upstream_version",
        "new_upstream_version",
        "upstream_branch",
        "upstream_branch_browse",
        "upstream_revisions",
        "imported_revisions",
        "include_upstream_history",
    ]

    def __init__(
        self,
        include_upstream_history,
        old_upstream_version,
        new_upstream_version,
        upstream_branch,
        upstream_branch_browse,
        upstream_revisions,
        imported_revisions,
    ):
        self.old_upstream_version = old_upstream_version
        self.new_upstream_version = new_upstream_version
        self.upstream_branch = upstream_branch
        self.upstream_branch_browse = upstream_branch_browse
        self.upstream_revisions = upstream_revisions
        self.imported_revisions = imported_revisions
        self.include_upstream_history = include_upstream_history


def detect_include_upstream_history(
    packaging_branch, upstream_branch_source, package, old_upstream_version
):
    # Simple heuristic: Find the old upstream version and see if it's present
    # in the history of the packaging branch
    try:
        (revision, _subpath) = upstream_branch_source.version_as_revision(
            package, old_upstream_version
        )
    except PackageVersionNotPresent:
        logging.warning(
            "Old upstream version %r is not present in upstream "
            "branch %r. Unable to determine whether upstream history "
            "is normally included. Assuming no.",
            old_upstream_version,
            upstream_branch_source,
        )
        return False

    graph = packaging_branch.repository.get_graph()
    ret = graph.is_ancestor(revision, packaging_branch.last_revision())
    if ret:
        logging.info(
            "Including upstream history, since previous upstream version "
            "(%s) is present in packaging branch history.",
            old_upstream_version,
        )
    else:
        logging.info(
            "Not including upstream history, since previous upstream version "
            "(%s) is not present in packaging branch history.",
            old_upstream_version,
        )

    return ret


class BranchOpenError(Exception):
    """Generic exception for branch open failures."""


def _convert_exception(url: str, e: Exception) -> Optional[BranchOpenError]:
    if isinstance(e, socket.error):
        return BranchOpenError(url, "Socket error: %s" % e)
    if isinstance(e, errors.NotBranchError):
        return BranchOpenError(url, "Branch does not exist: %s" % e)
    if isinstance(e, UnsupportedProtocol):
        return BranchOpenError(url, str(e))
    if isinstance(e, errors.ConnectionError):
        return BranchOpenError(url, str(e))
    if isinstance(e, errors.PermissionDenied):
        return BranchOpenError(url, str(e))
    if isinstance(e, errors.InvalidHttpResponse):
        if "Unexpected HTTP status 429" in str(e):
            raise BranchOpenError(url, str(e))
        return BranchOpenError(url, str(e))
    if isinstance(e, errors.TransportError):
        return BranchOpenError(url, str(e))
    if isinstance(e, UnusableRedirect):
        return BranchOpenError(url, str(e))
    if isinstance(e, errors.UnsupportedFormatError):
        return BranchOpenError(url, str(e))
    if isinstance(e, errors.UnknownFormatError):
        return BranchOpenError(url, str(e))
    if isinstance(e, RemoteGitError):
        return BranchOpenError(url, str(e))
    if isinstance(e, LineEndingError):
        return BranchOpenError(url, str(e))
    return None


def open_branch(
    url: str,
    possible_transports: Optional[list[Transport]] = None,
    name: str = None,
) -> Branch:
    """Open a branch by URL."""
    from breezy.directory_service import directories
    url = directories.dereference(url, purpose='read')

    url, params = urlutils.split_segment_parameters(url)
    if name is None:
        try:
            name = urlutils.unquote(params["branch"])
        except KeyError:
            name = None
    try:
        transport = get_transport(url, possible_transports=possible_transports)
        dir = ControlDir.open_from_transport(transport)
        return dir.open_branch(name=name)
    except Exception as e:
        converted = _convert_exception(url, e)
        if converted is not None:
            raise converted
        raise e


def find_new_upstream(  # noqa: C901
    tree,
    subpath: str,
    config,
    package,
    location: Optional[str] = None,
    old_upstream_version=None,
    new_upstream_version=None,
    trust_package: bool = False,
    version_kind: str = "release",
    allow_ignore_upstream_branch: bool = False,
    top_level: bool = False,
    create_dist=None,
    include_upstream_history: Optional[bool] = None,
    force_big_version_jump: bool = False,
    require_uscan: bool = False,
    skip_signatures: bool = False,
):

    # TODO(jelmer): Find the lastest upstream present in the upstream branch
    # rather than what's in the changelog.

    upstream_branch_location, upstream_branch_browse = (
            get_upstream_branch_location(
                tree, subpath, config, trust_package=trust_package))

    if upstream_branch_location:
        (upstream_branch_url, upstream_branch_name,
         upstream_subpath) = split_vcs_url(upstream_branch_location)
        if upstream_branch_name:
            upstream_branch_url = urlutils.join_segment_parameters(
                upstream_branch_url, {'branch': upstream_branch_name})
        try:
            upstream_branch = open_branch(upstream_branch_url)
        except BranchOpenError as e:
            if version_kind != "snapshot" and allow_ignore_upstream_branch:
                logging.warning(
                    "Upstream branch %s inaccessible; ignoring. %s",
                    upstream_branch_location,
                    e,
                )
            else:
                raise UpstreamBranchUnavailable(upstream_branch_location, e)
            upstream_branch = None
            upstream_branch_browse = None
        except urlutils.InvalidURL as e:
            if version_kind != "snapshot" and allow_ignore_upstream_branch:
                logging.warning(
                    "Upstream branch location %s invalid; ignoring. %s",
                    e.path, e)
            else:
                raise UpstreamBranchLocationInvalid(e.path, e.extra)
            upstream_branch = None
            upstream_branch_browse = None
    else:
        upstream_branch = None
        upstream_subpath = None  # noqa: F841

    if upstream_branch is not None:
        try:
            upstream_branch_source = UpstreamBranchSource.from_branch(
                upstream_branch,
                config=config,
                local_dir=tree.controldir,
                create_dist=create_dist,
                version_kind=version_kind,
                subpath=upstream_subpath,
            )
        except InvalidHttpResponse as e:
            raise UpstreamBranchUnavailable(upstream_branch_location, str(e))
        except ssl.SSLError as e:
            raise UpstreamBranchUnavailable(upstream_branch_location, str(e))
    else:
        upstream_branch_source = None

    if location is not None:
        try:
            branch = open_branch(location)
        except BranchOpenError:
            primary_upstream_source = TarfileSource(
                location, new_upstream_version)
        else:
            primary_upstream_source = UpstreamBranchSource.from_branch(
                branch,
                config=config,
                local_dir=tree.controldir,
                create_dist=create_dist,
                version_kind=version_kind
            )
    else:
        if version_kind == "snapshot":
            if upstream_branch_source is None:
                raise UpstreamBranchUnknown()
            primary_upstream_source = upstream_branch_source
        else:
            try:
                primary_upstream_source = UScanSource.from_tree(
                    tree, subpath, top_level, auto_fix=True,
                    skip_signatures=skip_signatures,
                )
            except NoWatchFile:
                # TODO(jelmer): Call out to lintian_brush.watch to generate a
                # watch file.
                if upstream_branch_source is None:
                    raise NoUpstreamLocationsKnown(package)
                if require_uscan:
                    raise
                primary_upstream_source = upstream_branch_source

    if new_upstream_version is None and primary_upstream_source is not None:
        unmangled_new_upstream_version, new_upstream_version = (
                primary_upstream_source.get_latest_version(
                    package, old_upstream_version))
    else:
        new_upstream_version = debianize_upstream_version(
            new_upstream_version)

    if new_upstream_version is None:
        raise NewUpstreamMissing()

    if not new_upstream_version[0].isdigit():
        # dpkg forbids this, let's just refuse it early
        raise InvalidFormatUpstreamVersion(
            new_upstream_version, primary_upstream_source
        )

    try:
        new_upstream_version = Version(new_upstream_version)
    except ValueError:
        raise InvalidFormatUpstreamVersion(
            new_upstream_version, primary_upstream_source
        )

    if old_upstream_version:
        if (strip_dfsg_suffix(str(old_upstream_version))
                == strip_dfsg_suffix(str(new_upstream_version))):
            raise UpstreamAlreadyImported(new_upstream_version)
        if old_upstream_version > new_upstream_version:
            if version_kind == "release" and matches_release(
                str(old_upstream_version), str(new_upstream_version)
            ):
                raise UpstreamAlreadyImported(new_upstream_version)
            raise NewerUpstreamAlreadyImported(
                old_upstream_version, new_upstream_version
            )
        if (is_big_version_jump(old_upstream_version, new_upstream_version) and
                not force_big_version_jump):
            raise BigVersionJump(old_upstream_version, new_upstream_version)

    # TODO(jelmer): Check if new_upstream_version is already imported

    logging.info("Using version string %s.", new_upstream_version)

    if include_upstream_history is None and upstream_branch_source is not None:
        include_upstream_history = detect_include_upstream_history(
            tree.branch, upstream_branch_source, package, old_upstream_version
        )

    if include_upstream_history is False:
        upstream_branch_source = None

    # Look up the revision id from the version string
    if upstream_branch_source is not None:
        try:
            upstream_revisions = upstream_branch_source.version_as_revisions(
                package, str(new_upstream_version)
            )
        except PackageVersionNotPresent:
            if upstream_branch_source is primary_upstream_source:
                # The branch is our primary upstream source, so if it can't
                # find the version then there's nothing we can do.
                raise UpstreamVersionMissingInUpstreamBranch(
                    upstream_branch_source.upstream_branch,
                    str(new_upstream_version)
                )
            elif not allow_ignore_upstream_branch:
                raise UpstreamVersionMissingInUpstreamBranch(
                    upstream_branch_source.upstream_branch,
                    str(new_upstream_version)
                )
            else:
                logging.warning(
                    "Upstream version %s is not in upstream branch %s. "
                    "Not merging from upstream branch. ",
                    new_upstream_version,
                    upstream_branch_source.upstream_branch,
                )
                upstream_revisions = None
                upstream_branch_source = None
    else:
        upstream_revisions = None

    try:
        files_excluded = get_files_excluded(tree, subpath, top_level)
    except NoSuchFile:
        files_excluded = None
    else:
        if files_excluded:
            new_upstream_version = Version(add_dfsg_suffix(
                str(new_upstream_version),
                old_upstream_version))
            logging.info(
                "Adding DFSG suffix since upstream files are excluded: %s",
                new_upstream_version)

    return (
        primary_upstream_source,
        new_upstream_version,
        upstream_revisions,
        upstream_branch_source,
        upstream_branch,
        upstream_branch_browse,
        files_excluded,
        include_upstream_history,
    )


def import_upstream(
    tree: Tree,
    version_kind: str = "release",
    location: Optional[str] = None,
    new_upstream_version: Optional[Union[Version, str]] = None,
    force: bool = False,
    distribution_name: str = DEFAULT_DISTRIBUTION,
    allow_ignore_upstream_branch: bool = True,
    trust_package: bool = False,
    committer: Optional[str] = None,
    subpath: str = "",
    include_upstream_history: Optional[bool] = None,
    create_dist:
        Optional[Callable[[Tree, str, Version, str], Optional[str]]] = None,
    force_big_version_jump: bool = False,
    skip_signatures: bool = False,
) -> ImportUpstreamResult:
    """Import a new upstream version into a tree.

    Raises:
      InvalidFormatUpstreamVersion
      PreviousVersionTagMissing
      UnsupportedRepackFormat
      DistCommandFailed
      MissingChangelogError
      MissingUpstreamTarball
      NewUpstreamMissing
      UpstreamBranchUnavailable
      UpstreamAlreadyImported
      QuiltError
      UpstreamVersionMissingInUpstreamBranch
      UpstreamBranchUnknown
      PackageIsNative
      InconsistentSourceFormatError
      ChangelogParseError
      UScanError
      UpstreamMetadataSyntaxError
      UpstreamNotBundled
      NewUpstreamTarballMissing
      NoUpstreamLocationsKnown
      NewerUpstreamAlreadyImported
    Returns:
      ImportUpstreamResult object
    """
    # TODO(jelmer): support more components
    components = [None]
    config = debuild_config(tree, subpath)
    (changelog, top_level) = find_changelog(
        tree, subpath, merge=False, max_blocks=2)
    old_upstream_version = changelog.version.upstream_version
    package = changelog.package
    contains_upstream_source = tree_contains_upstream_source(tree, subpath)
    build_type = config.build_type
    if build_type is None:
        build_type = guess_build_type(
            tree,
            changelog.version,
            subpath,
            contains_upstream_source=contains_upstream_source,
        )
    if build_type == BUILD_TYPE_MERGE:
        raise UpstreamNotBundled(changelog.package)
    if build_type == BUILD_TYPE_NATIVE:
        raise PackageIsNative(changelog.package, changelog.version)

    (
        primary_upstream_source,
        new_upstream_version,
        upstream_revisions,
        upstream_branch_source,
        upstream_branch,
        upstream_branch_browse,
        files_excluded,
        include_upstream_history,
    ) = find_new_upstream(
        tree,
        subpath,
        config,
        package,
        location=location,
        old_upstream_version=old_upstream_version,
        new_upstream_version=new_upstream_version,
        trust_package=trust_package,
        version_kind=version_kind,
        allow_ignore_upstream_branch=allow_ignore_upstream_branch,
        top_level=top_level,
        include_upstream_history=include_upstream_history,
        create_dist=create_dist,
        force_big_version_jump=force_big_version_jump,
        skip_signatures=skip_signatures,
    )

    with tempfile.TemporaryDirectory() as target_dir:
        initial_path = os.path.join(target_dir, "initial")
        os.mkdir(initial_path)
        try:
            locations = primary_upstream_source.fetch_tarballs(
                package, str(new_upstream_version), target_dir,
                components=components)
        except (PackageVersionNotPresent, WatchLineWithoutMatchingHrefs):
            if upstream_revisions is not None:
                locations = upstream_branch_source.fetch_tarballs(
                    package,
                    str(new_upstream_version),
                    initial_path,
                    components=components,
                    revisions=upstream_revisions)
            else:
                raise
        orig_path = os.path.join(target_dir, "orig")
        try:
            tarball_filenames = get_tarballs(
                orig_path, tree, package, new_upstream_version, locations
            )
        except FileExists as e:
            raise AssertionError(
                "The target file %s already exists, and is either "
                "different to the new upstream tarball, or they "
                "are of different formats. Either delete the target "
                "file, or use it as the argument to import." % e.path
            )
        imported_revisions = do_import(
            tree,
            subpath,
            tarball_filenames,
            package,
            str(new_upstream_version),
            str(old_upstream_version),
            upstream_branch_source.upstream_branch
            if upstream_branch_source else None,
            upstream_revisions,
            merge_type=None,
            force=force,
            committer=committer,
            files_excluded=files_excluded,
        )

    return ImportUpstreamResult(
        include_upstream_history=include_upstream_history,
        old_upstream_version=old_upstream_version,
        new_upstream_version=new_upstream_version,
        upstream_branch=upstream_branch,
        upstream_branch_browse=upstream_branch_browse,
        upstream_revisions=upstream_revisions,
        imported_revisions=imported_revisions,
    )


class MergeUpstreamResult:
    """Object representing the result of a merge_upstream operation."""

    __slots__ = [
        "old_upstream_version",
        "new_upstream_version",
        "upstream_branch",
        "upstream_branch_browse",
        "upstream_revisions",
        "old_revision",
        "new_revision",
        "imported_revisions",
        "include_upstream_history",
    ]

    def __init__(
        self,
        include_upstream_history,
        old_upstream_version,
        new_upstream_version,
        upstream_branch,
        upstream_branch_browse,
        upstream_revisions,
        old_revision,
        new_revision,
        imported_revisions,
    ):
        self.include_upstream_history = include_upstream_history
        self.old_upstream_version = old_upstream_version
        self.new_upstream_version = new_upstream_version
        self.upstream_branch = upstream_branch
        self.upstream_branch_browse = upstream_branch_browse
        self.upstream_revisions = upstream_revisions
        self.old_revision = old_revision
        self.new_revision = new_revision
        self.imported_revisions = imported_revisions

    def __tuple__(self):
        # Backwards compatibility
        return (self.old_upstream_version, self.new_upstream_version)


def merge_upstream(  # noqa: C901
    tree: Tree,
    version_kind: str = "release",
    location: Optional[str] = None,
    new_upstream_version: Optional[str] = None,
    *,
    force: bool = False,
    distribution_name: str = DEFAULT_DISTRIBUTION,
    allow_ignore_upstream_branch: bool = True,
    trust_package: bool = False,
    committer: Optional[str] = None,
    update_changelog: bool = True,
    subpath: str = "",
    include_upstream_history: Optional[bool] = None,
    create_dist: Optional[
        Callable[[Tree, str, Version, str], Optional[str]]] = None,
    force_big_version_jump: bool = False,
    debian_revision: Optional[str] = None,
    require_uscan: bool = False,
    skip_signatures: bool = False,
    skip_empty: bool = False,
) -> MergeUpstreamResult:
    """Merge a new upstream version into a tree.

    Raises:
      InvalidFormatUpstreamVersion
      PreviousVersionTagMissing
      UnsupportedRepackFormat
      DistCommandFailed
      MissingChangelogError
      MissingUpstreamTarball
      NewUpstreamMissing
      UpstreamBranchUnavailable
      UpstreamAlreadyMerged
      UpstreamAlreadyImported
      UpstreamMergeConflicted
      QuiltError
      UpstreamVersionMissingInUpstreamBranch
      UpstreamBranchUnknown
      PackageIsNative
      InconsistentSourceFormatError
      ChangelogParseError
      UScanError
      NoUpstreamLocationsKnown
      UpstreamMetadataSyntaxError
      NewerUpstreamAlreadyImported
      WatchSyntaxError
      ReleaseWithoutChanges
    Returns:
      MergeUpstreamResult object
    """
    components = [None]
    config = debuild_config(tree, subpath)
    (changelog, top_level) = find_changelog(
        tree, subpath, merge=False, max_blocks=2)
    old_upstream_version = changelog.version.upstream_version
    old_revision = tree.last_revision()
    package = changelog.package
    contains_upstream_source = tree_contains_upstream_source(tree, subpath)
    build_type = config.build_type
    if build_type is None:
        build_type = guess_build_type(
            tree,
            changelog.version,
            subpath,
            contains_upstream_source=contains_upstream_source,
        )
    need_upstream_tarball = build_type != BUILD_TYPE_MERGE
    if build_type == BUILD_TYPE_NATIVE:
        raise PackageIsNative(changelog.package, changelog.version)

    (
        primary_upstream_source,
        new_upstream_version,
        upstream_revisions,
        upstream_branch_source,
        upstream_branch,
        upstream_branch_browse,
        files_excluded,
        include_upstream_history,
    ) = find_new_upstream(
        tree,
        subpath,
        config,
        package,
        location=location,
        old_upstream_version=old_upstream_version,
        new_upstream_version=new_upstream_version,
        trust_package=trust_package,
        version_kind=version_kind,
        allow_ignore_upstream_branch=allow_ignore_upstream_branch,
        top_level=top_level,
        include_upstream_history=include_upstream_history,
        create_dist=create_dist,
        force_big_version_jump=force_big_version_jump,
        require_uscan=require_uscan,
        skip_signatures=skip_signatures,
    )

    if need_upstream_tarball:
        with tempfile.TemporaryDirectory() as target_dir:
            initial_path = os.path.join(target_dir, "initial")
            os.mkdir(initial_path)

            try:
                locations = primary_upstream_source.fetch_tarballs(
                    package, str(new_upstream_version), initial_path,
                    components=components)
            except PackageVersionNotPresent as e:
                if upstream_revisions is not None:
                    locations = upstream_branch_source.fetch_tarballs(
                        package,
                        str(new_upstream_version),
                        initial_path,
                        components=components,
                        revisions=upstream_revisions,
                    )
                else:
                    raise NewUpstreamTarballMissing(
                        e.package, e.version, e.upstream)

            orig_path = os.path.join(target_dir, "orig")
            os.mkdir(orig_path)
            try:
                tarball_filenames = get_tarballs(
                    orig_path, tree, package, new_upstream_version, locations
                )
            except FileExists as e:
                raise AssertionError(
                    "The target file %s already exists, and is either "
                    "different to the new upstream tarball, or they "
                    "are of different formats. Either delete the target "
                    "file, or use it as the argument to import." % e.path
                )
            try:
                conflicts, imported_revids = do_merge(
                    tree,
                    tarball_filenames,
                    package,
                    str(new_upstream_version),
                    str(old_upstream_version),
                    upstream_branch_source.upstream_branch
                    if upstream_branch_source
                    else None,
                    upstream_revisions,
                    merge_type=None,
                    force=force,
                    committer=committer,
                    files_excluded=files_excluded,
                )
            except UpstreamBranchAlreadyMerged:
                # TODO(jelmer): Perhaps reconcile these two exceptions?
                raise UpstreamAlreadyMerged(new_upstream_version)
            except UpstreamAlreadyImported:
                pristine_tar_source = get_pristine_tar_source(
                    tree, tree.branch)
                imported_revids = get_existing_imported_upstream_revids(
                    pristine_tar_source, package, new_upstream_version)
                for entry in imported_revids:
                    if entry[0] is None:
                        upstream_revid = entry[2]
                        break
                else:
                    upstream_revid = None
                try:
                    try:
                        conflicts = tree.merge_from_branch(
                            pristine_tar_source.branch,
                            to_revision=upstream_revid)
                    except NoCommits:
                        # Maybe it's somewhere in our history, but
                        # there isn't a viable upstream branch.
                        conflicts = tree.merge_from_branch(
                            tree.branch, to_revision=upstream_revid
                        )
                except PointlessMerge:
                    raise UpstreamAlreadyMerged(new_upstream_version)
    else:
        conflicts = 0
        imported_revids = []

    if skip_empty:
        changes = tree.iter_changes(tree.basis_tree())
        try:
            next(changes)
        except StopIteration:
            tree.set_pending_merges([])
            raise ReleaseWithoutChanges(new_upstream_version)

    # Re-read changelog, since it may have been changed by the merge
    # from upstream.
    try:
        (changelog, top_level) = find_changelog(
            tree, subpath, False, max_blocks=2)
    except (ChangelogParseError, MissingChangelogError):
        # If there was a conflict that affected debian/changelog, then that
        # might be to blame.
        if conflicts:
            raise UpstreamMergeConflicted(old_upstream_version, conflicts)
        raise
    if top_level:
        debian_path = subpath
    else:
        debian_path = os.path.join(subpath, "debian")

    if Version(old_upstream_version) >= Version(new_upstream_version):
        if conflicts:
            raise UpstreamMergeConflicted(old_upstream_version, conflicts)
        raise UpstreamAlreadyMerged(new_upstream_version)

    try:
        with ChangelogEditor(
                tree.abspath(os.path.join(debian_path, "changelog"))) as cl:
            if debian_revision is None:
                debian_revision = initial_debian_revision(distribution_name)
            new_version = str(new_upstream_package_version(
                new_upstream_version, debian_revision, cl[0].version.epoch))
            if not update_changelog:
                # We need to run "gbp dch" here, since the next "gbp dch" runs
                # won't pick up the pending changes, as we're about to change
                # debian/changelog.
                from debmutate.changelog import gbp_dch
                gbp_dch(tree.basedir)
            cl.auto_version(new_version)

            cl.add_entry([upstream_merge_changelog_line(new_upstream_version)])
    except GeneratedFile as e:
        raise ChangelogGeneratedFile(e.path, e.template_path, e.template_type)

    if not need_upstream_tarball:
        logging.info("The changelog has been updated for the new version.")
    else:
        if conflicts:
            raise UpstreamMergeConflicted(new_upstream_version, conflicts)

    debcommit(tree, subpath=subpath, committer=committer)

    return MergeUpstreamResult(
        include_upstream_history=include_upstream_history,
        old_upstream_version=old_upstream_version,
        new_upstream_version=new_upstream_version,
        old_revision=old_revision,
        new_revision=tree.last_revision(),
        upstream_branch=upstream_branch,
        upstream_branch_browse=upstream_branch_browse,
        upstream_revisions=upstream_revisions,
        imported_revisions=imported_revids,
    )


def report_fatal(
        code: str, description: str, *, upstream_version: Optional[str] = None,
        details=None, hint: Optional[str] = None,
        transient: Optional[bool] = None, stage=None):
    if os.environ.get('SVP_API') == '1':
        context = {}
        if upstream_version is not None:
            context['upstream_version'] = str(upstream_version)
        with open(os.environ['SVP_RESULT'], 'w') as f:
            json.dump({
                'result_code': code,
                'hint': hint,
                'transient': transient,
                'stage': stage,
                'description': description,
                'versions': versions_dict(),
                'details': details,
                'context': context}, f)
    logging.fatal('%s', description)
    if hint:
        logging.info('%s', hint)


def main(argv=None):
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--trust-package",
        action="store_true",
        default=False,
        help=argparse.SUPPRESS,
    )
    parser.add_argument(
        "--import-only",
        action="store_true",
        help="Only import a new version, do not merge.",
    )
    parser.add_argument(
        "--update-packaging",
        action="store_true",
        default=False,
        help="Attempt to update packaging to upstream changes.",
    )
    parser.add_argument(
        "--snapshot",
        help="Merge a new upstream snapshot rather than a release",
        action="store_true",
    )
    parser.add_argument(
        "--refresh-patches",
        action="store_true",
        help="Refresh quilt patches after upstream merge.",
    )
    parser.add_argument(
        "--dist-command",
        type=str,
        help="Command to run to create tarball from source tree.",
        default=os.environ.get('DIST'),
    )
    parser.add_argument(
        "--no-include-upstream-history",
        action="store_false",
        default=None,
        dest="include_upstream_history",
        help="do not include upstream branch history",
    )
    parser.add_argument(
        "--include-upstream-history",
        action="store_true",
        dest="include_upstream_history",
        help="force inclusion of upstream history",
        default=None,
    )
    parser.add_argument(
        "--force-big-version-jump",
        action="store_true",
        help="force through big version jumps")
    parser.add_argument(
        "--debian-revision",
        type=str,
        help="Debian revision to use (e.g. '1' or '0ubuntu1')",
        default=None)
    parser.add_argument(
        "--require-uscan",
        action="store_true",
        help=("Require that uscan provides a tarball"
              "(if --snapshot is not specified)"))
    parser.add_argument(
        "--upstream-location",
        type=str,
        default=None,
        help="Location of the upstream location "
        "(defaults to reading debian/upstream/metadata or guessing)")
    parser.add_argument(
        "--verbose", action="store_true")
    parser.add_argument(
        "--skip-signatures", action="store_true",
        help="Skip signature validation")
    parser.add_argument(
        "--skip-empty", action="store_true",
        help="Skip releases without changes")
    # TODO(jelmer): Support "auto"
    parser.add_argument(
        "--version-kind", choices=["snapshot", "release"],
        default="release",
        help="Version kind to merge.")
    args = parser.parse_args(argv)

    if args.verbose:
        loglevel = logging.DEBUG
    else:
        loglevel = logging.INFO
    logging.basicConfig(level=loglevel, format='%(message)s')

    from breezy.plugin import load_plugins
    load_plugins()

    local_tree, subpath = WorkingTree.open_containing('.')

    with local_tree.lock_write():
        committer = os.environ.get('COMMITTER')

        if control_files_in_root(local_tree, subpath):
            report_fatal(
                "control-files-in-root",
                "control files live in root rather than debian/ "
                "(LarstIQ mode)",
                transient=False,
            )
            return 1

        if args.dist_command:

            def create_dist(tree, package, version, target_dir, subpath=''):
                try:
                    return run_dist_command(
                        tree, package, version, target_dir, args.dist_command,
                        subpath=subpath)
                except MissingNestedTree as e:
                    raise DistMissingNestedTree(e.path) from e
        else:
            create_dist = None

        update_changelog = True
        if os.environ.get('DEB_UPDATE_CHANGELOG') == 'leave':
            update_changelog = False
        elif os.environ.get('DEB_UPDATE_CHANGELOG') == 'update':
            update_changelog = True

        if args.snapshot:
            version_kind = "snapshot"
        else:
            version_kind = args.version_kind

        try:
            if not args.import_only:
                try:
                    result = merge_upstream(
                        tree=local_tree,
                        version_kind=version_kind,
                        trust_package=args.trust_package,
                        update_changelog=update_changelog,
                        subpath=subpath,
                        committer=committer,
                        include_upstream_history=args.include_upstream_history,
                        create_dist=create_dist,
                        force_big_version_jump=args.force_big_version_jump,
                        debian_revision=args.debian_revision,
                        require_uscan=args.require_uscan,
                        location=args.upstream_location,
                        skip_signatures=args.skip_signatures,
                        skip_empty=args.skip_empty,
                    )
                except MalformedTransform:
                    traceback.print_exc()
                    error_description = (
                        "Malformed tree transform during new upstream merge"
                    )
                    error_code = "malformed-transform"
                    report_fatal(
                        error_code, error_description, transient=False)
                    return 1
                except CorruptUpstreamSourceFile as e:
                    report_fatal(
                        'corrupt-upstream-file',
                        str(e), transient=False)
                except UncommittedChanges as e:
                    report_fatal(
                        'uncommitted-changes', str(e), transient=False)
                    return 1
            else:
                result = import_upstream(
                    tree=local_tree,
                    version_kind=version_kind,
                    trust_package=args.trust_package,
                    subpath=subpath,
                    committer=committer,
                    include_upstream_history=args.include_upstream_history,
                    create_dist=create_dist,
                    force_big_version_jump=args.force_big_version_jump,
                    skip_signatures=args.skip_signatures,
                )
        except MemoryError as e:
            report_fatal("memory-error", str(e))
            return 1
        except UpstreamAlreadyImported as e:
            if version_kind == "snapshot":
                report_fatal(
                    "nothing-to-do",
                    "Last upstream version %s already imported." % e.version,
                    upstream_version=e.version,
                    transient=False
                )
            else:
                report_fatal(
                    "nothing-to-do",
                    "Last upstream version %s already imported. " % e.version,
                    upstream_version=e.version,
                    hint="Import a snapshot by specifying --snapshot.",
                    transient=False
                )
            return 1
        except ReleaseWithoutChanges as e:
            report_fatal(
                "nothing-to-do",
                "New release %s is available, but does not contain changes."
                % e.new_upstream_version,
                upstream_version=e.new_upstream_version,
                transient=False
            )
            return 1
        except UpstreamBranchLocationInvalid as e:
            report_fatal(
                "upstream-branch-invalid",
                "The upstream branch location ({}) is invalid: {}".format(
                    e.url, e.extra), transient=False)
            return 1
        except ChangelogGeneratedFile as e:
            report_fatal(
                "changelog-generated-file",
                "changelog file can't be updated because it is generated. "
                "(template type: %s, path: %s)" % (
                    e.template_type, e.template_path))
            return 1
        except UnsupportedRepackFormat as e:
            error_description = (
                "Unable to repack file %s to supported tarball format."
                % (os.path.basename(e.location))
            )
            report_fatal(
                "unsupported-repack-format", error_description,
                transient=False)
            return 1
        except NewUpstreamMissing:
            report_fatal(
                "new-upstream-missing",
                "Unable to find new upstream source.")
            return 1
        except UpstreamAlreadyMerged as e:
            report_fatal(
                "nothing-to-do",
                "Last upstream version %s already merged." % e.version,
                upstream_version=e.version, transient=False)
            return 1
        except NoWatchFile:
            report_fatal(
                "no-watch-file",
                "No watch file is present, but --require-uscan was specified",
                transient=False)
            return 1
        except PreviousVersionTagMissing as e:
            report_fatal(
                "previous-upstream-missing",
                "Previous upstream version %s missing (tag: %s)."
                % (e.version, e.tag_name), transient=False
            )
            return 1
        except InvalidFormatUpstreamVersion as e:
            report_fatal(
                "invalid-upstream-version-format",
                "{!r} reported invalid format version string {}.".format(
                    e.source, e.version), transient=False
            )
            return 1
        except PristineTarError as e:
            report_fatal("pristine-tar-error", "Pristine tar error: %s" % e)
            return 1
        except UpstreamBranchUnavailable as e:
            error_description = (
                "The upstream branch at {} was unavailable: {}".format(
                    e.location, e.error))
            transient = None
            error_code = "upstream-branch-unavailable"
            if "Fossil branches are not yet supported" in str(e.error):
                error_code = "upstream-unsupported-vcs-fossil"
                transient = False
            if "Mercurial branches are not yet supported." in str(e.error):
                error_code = "upstream-unsupported-vcs-hg"
                transient = False
            if "Subversion branches are not yet supported." in str(e.error):
                error_code = "upstream-unsupported-vcs-svn"
                transient = False
            if "Darcs branches are not yet supported" in str(e.error):
                error_code = "upstream-unsupported-vcs-darcs"
                transient = False
            if "Unsupported protocol for url" in str(e.error):
                transient = False
                if "svn://" in str(e.error):
                    error_code = "upstream-unsupported-vcs-svn"
                elif "cvs+pserver://" in str(e.error):
                    error_code = "upstream-unsupported-vcs-cvs"
                else:
                    error_code = "upstream-unsupported-vcs"
            report_fatal(error_code, error_description, transient=transient)
            return 1
        except UpstreamBranchUnknown:
            report_fatal(
                "upstream-branch-unknown",
                "Upstream branch location unknown.",
                hint="Set 'Repository' field in debian/upstream/metadata?",
                transient=False
            )
            return 1
        except UpstreamMergeConflicted as e:
            if isinstance(e.conflicts, int):
                conflicts = e.conflicts
            else:
                conflicts = [[c.path, c.typestring] for c in e.conflicts]
            report_fatal(
                "upstream-merged-conflicts",
                "Merging upstream version %s resulted in conflicts."
                % e.version,
                upstream_version=e.version, details={'conflicts': conflicts},
                transient=False)
            return 1
        except PackageIsNative as e:
            report_fatal(
                "native-package",
                "Package {} is native; unable to merge new upstream.".format(
                    e.package),
                transient=False
            )
            return 1
        except ChangelogParseError as e:
            error_description = str(e)
            error_code = "unparseable-changelog"
            report_fatal(error_code, error_description, transient=False)
            return 1
        except UpstreamVersionMissingInUpstreamBranch as e:
            error_description = (
                "Upstream version {} not in upstream branch {!r}".format(
                    e.version, e.branch))
            error_code = "upstream-version-missing-in-upstream-branch"
            report_fatal(error_code, error_description, transient=False)
            return 1
        except InconsistentSourceFormatError as e:
            report_fatal(
                "inconsistent-source-format",
                "Inconsistencies in type of package: %s" % e,
                transient=False
            )
            return 1
        except WatchLineWithoutMatches as e:
            report_fatal(
                "uscan-watch-line-without-matches",
                "UScan did not find matches for line: %s" % e.line.strip(),
                transient=False
            )
            return 1
        except NoRoundtrippingSupport:
            error_description = (
                "Unable to import upstream repository into "
                "packaging repository."
            )
            error_code = "roundtripping-error"
            report_fatal(error_code, error_description, transient=False)
            return 1
        except UScanError as e:
            error_description = str(e)
            if e.errors == "OpenPGP signature did not verify.":
                error_code = "upstream-pgp-signature-verification-failed"
            else:
                error_code = "uscan-error"
            report_fatal(error_code, error_description)
            return 1
        except UpstreamMetadataSyntaxError as e:
            report_fatal(
                "upstream-metadata-syntax-error",
                "Unable to parse {}: {}".format(
                    e.path, e.error), transient=False)
            return 1
        except InvalidNormalization as e:
            error_description = str(e)
            error_code = "invalid-path-normalization"
            report_fatal(error_code, error_description, transient=False)
            return 1
        except MissingChangelogError as e:
            report_fatal(
                "missing-changelog", "Missing changelog %s" % e,
                transient=False)
            return 1
        except DistCommandFailed as e:
            report_fatal(
                e.kind or "dist-command-failed", e.error, transient=False,
                stage=("dist", ))
            return 1
        except MissingUpstreamTarball as e:
            report_fatal(
                "missing-upstream-tarball",
                "Missing upstream tarball: %s" % e
            )
            return 1
        except NewUpstreamTarballMissing as e:
            report_fatal(
                "new-upstream-tarball-missing",
                "New upstream version (%s/%s) found, but was missing "
                "when retrieved as tarball from %r."
                % (e.package, e.version, e.upstream),
                upstream_version=e.version)
            return 1
        except NoUpstreamLocationsKnown:
            report_fatal(
                "no-upstream-locations-known",
                "No debian/watch file or Repository in "
                "debian/upstream/metadata to retrieve new upstream version "
                "from.",
            )
            return 1
        except NewerUpstreamAlreadyImported as e:
            report_fatal(
                "newer-upstream-version-already-imported",
                "A newer upstream release (%s) has already been imported. "
                "Found: %s" % (e.old_upstream_version, e.new_upstream_version),
                upstream_version=e.new_upstream_version, transient=False
            )
            return 1
        except MergingUpstreamSubpathUnsupported as e:
            report_fatal(
                "unsupported-merging-upstream-with-subpath",
                str(e), transient=False)
        except WatchSyntaxError as e:
            report_fatal('watch-syntax-error', str(e), transient=False)
            return 1
        except BigVersionJump as e:
            report_fatal(
                "big-version-jump",
                "There was a big jump in upstream versions: {} ⇒ {}".format(
                    e.old_upstream_version, e.new_upstream_version),
                upstream_version=e.new_upstream_version, transient=False)
            return 1
        except DistMissingNestedTree as e:
            report_fatal(
                "requires-nested-tree-support",
                "Unable to find nested tree at %s" % e.path,
                transient=False, stage=("dist", ))
            return 1
        except OSError as e:
            if e.errno == errno.ENOSPC:
                report_fatal("no-space-on-device", str(e))
                return 1
            raise

        svp_context = {
            "old_upstream_version":
                str(result.old_upstream_version)
                if result.old_upstream_version else None,
            "upstream_version": str(result.new_upstream_version),
        }

        if result.upstream_branch:
            svp_context["upstream_branch_url"] = full_branch_url(
                result.upstream_branch)
            svp_context["upstream_branch_browse"] = (
                result.upstream_branch_browse)

        svp_context["include_upstream_history"] = (
            result.include_upstream_history)

        if args.import_only:
            logging.info(
                "Imported new upstream version %s (previous: %s)",
                result.new_upstream_version,
                result.old_upstream_version,
            )

            if os.environ.get('SVP_API') == '1':
                with open(os.environ['SVP_RESULT'], 'w') as f:
                    json.dump({
                        'value': VALUE_IMPORT[version_kind],
                        'description': "Imported new upstream version %s" % (
                            result.new_upstream_version),
                        'context': svp_context,
                        'versions': versions_dict(),
                        }, f)
            return 0
        else:
            logging.info(
                "Merged new upstream version %s (previous: %s)",
                result.new_upstream_version,
                result.old_upstream_version,
            )

            if args.update_packaging:
                from .update_packaging import (
                    update_packaging,
                )
                old_tree = local_tree.branch.repository.revision_tree(
                    result.old_revision
                )
                notes = update_packaging(
                    local_tree, old_tree, committer=committer)
                svp_context["notes"] = notes
                for n in notes:
                    logging.info("%s", n)

            patch_series_path = os.path.join(subpath, "debian/patches/series")
            if args.refresh_patches and local_tree.has_filename(
                    patch_series_path):
                from .quilt_refresh import (
                    QuiltError,
                    QuiltPatchPushFailure,
                    QuiltPatchDoesNotApply,
                    refresh_quilt_patches,
                )
                logging.info("Refresh quilt patches.")
                try:
                    refresh_quilt_patches(
                        local_tree,
                        committer=committer,
                        subpath=subpath,
                    )
                except QuiltPatchDoesNotApply as e:
                    error_description = (
                        "Quilt patch %s no longer applies" % e.patch_name)
                    error_code = "quilt-patch-out-of-date"
                    report_fatal(error_code, error_description)
                    return 1
                except QuiltPatchPushFailure as e:
                    error_description = (
                        "An error occurred refreshing quilt patch %s"
                        % (e.patch_name, ))
                    error_code = "quilt-refresh-error"
                    report_fatal(error_code, error_description)
                    return 1
                except QuiltError as e:
                    error_description = (
                        "An error (%d) occurred refreshing quilt patches: "
                        "%s%s" % (e.retcode, e.stderr, e.extra)
                    )
                    error_code = "quilt-refresh-error"
                    report_fatal(
                        error_code, error_description, transient=False)
                    return 1
                except QuiltPatchPushFailure as e:
                    error_description = (
                        "An error occurred refreshing quilt patch %s: %s"
                        % (e.patch_name, e.actual_error.extra)
                    )
                    error_code = "quilt-refresh-error"
                    report_fatal(
                        error_code, error_description, transient=False)
                    return 1

            logging.info(
                'Merge new upstream version %s', result.new_upstream_version)
            proposed_commit_message = (
                "Merge new upstream release %s" % result.new_upstream_version
            )
            if os.environ.get('SVP_API') == '1':
                with open(os.environ['SVP_RESULT'], 'w') as f:
                    json.dump({
                        'value': VALUE_MERGE[version_kind],
                        'description': "Merged new upstream version %s" % (
                            result.new_upstream_version),
                        'commit-message': proposed_commit_message,
                        'context': svp_context,
                        'versions': versions_dict(),
                        }, f)


if __name__ == "__main__":
    import sys

    from breezy.commands import exception_to_return_code

    sys.exit(exception_to_return_code(main(sys.argv[1:])))



#!/usr/bin/python3
# Copyright (C) 2019 Jelmer Vernooij <jelmer@jelmer.uk>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

from contextlib import ExitStack
import json
import logging
import os
import sys
from urllib.parse import urlparse

import breezy.bzr  # noqa: F401
import breezy.git  # noqa: F401

from breezy import osutils
from breezy.branch import Branch
from breezy.workingtree import WorkingTree

from debmutate.changelog import ChangelogEditor
from debmutate.control import ControlEditor
from debmutate.deb822 import ChangeConflict
from debmutate.reformatting import GeneratedFile, FormattingUnpreservable

from breezy.plugins.debian.info import versions_dict
from breezy.plugins.debian.directory import vcs_git_url_to_bzr_url


BRANCH_NAME = "orphan"
QA_MAINTAINER = "Debian QA Group <packages@qa.debian.org>"


def push_to_salsa(local_tree, orig_branch, user, name, dry_run=False):
    from breezy import urlutils
    from breezy.branch import Branch
    from breezy.errors import PermissionDenied, AlreadyControlDirError
    from breezy.plugins.gitlab.forge import GitLab
    from breezy.forge import UnsupportedForge, get_forge, ForgeLoginRequired

    from silver_platter import pick_additional_colocated_branches
    from silver_platter.proposal import push_changes

    if dry_run:
        logging.info("Creating and pushing to salsa project %s/%s", user, name)
        return

    try:
        salsa = GitLab.probe_from_url("https://salsa.debian.org/")
    except ForgeLoginRequired:
        logging.warning("No login for salsa known, not pushing branch.")
        return

    try:
        orig_forge = get_forge(orig_branch)
    except UnsupportedForge:
        logging.debug("Original branch %r not hosted on salsa.")
        from_project = None
    else:
        if orig_forge == salsa:
            from_project = urlutils.URL.from_string(
                orig_branch.controldir.user_url).path
        else:
            from_project = None

    if from_project is not None:
        salsa.fork_project(from_project, owner=user)
    else:
        to_project = "{}/{}".format(user, name)
        try:
            salsa.create_project(to_project)
        except PermissionDenied as e:
            logging.info('No permission to create new project %s under %s: %s',
                         name, user, e)
            return
        except AlreadyControlDirError:
            logging.info('Project %s already exists, using..', to_project)
    target_branch = Branch.open(
        "git+ssh://git@salsa.debian.org/{}/{}.git".format(user, name)
    )
    additional_colocated_branches = pick_additional_colocated_branches(
        local_tree.branch
    )
    return push_changes(
        local_tree.branch,
        target_branch,
        forge=salsa,
        additional_colocated_branches=additional_colocated_branches,
        dry_run=dry_run,
    )


class OrphanResult:

    def __init__(
        self,
        package=None,
        old_vcs_url=None,
        new_vcs_url=None,
        salsa_user=None,
        wnpp_bug=None,
    ):
        self.package = package
        self.old_vcs_url = old_vcs_url
        self.new_vcs_url = new_vcs_url
        self.pushed = False
        self.salsa_user = salsa_user
        self.wnpp_bug = wnpp_bug

    def json(self):
        return {
            "package": self.package,
            "old_vcs_url": self.old_vcs_url,
            "new_vcs_url": self.new_vcs_url,
            "pushed": self.pushed,
            "salsa_user": self.salsa_user,
            "wnpp_bug": self.wnpp_bug,
        }


def connect_udd_mirror():
    import psycopg2

    return psycopg2.connect(
        database="udd",
        user="udd-mirror",
        password="udd-mirror",
        host="udd-mirror.debian.net",
    )


def find_wnpp_bug(source):
    conn = connect_udd_mirror()
    cursor = conn.cursor()
    cursor.execute(
        "SELECT id FROM wnpp WHERE type = 'O' AND source = %s", (source,))
    entry = cursor.fetchone()
    if entry is None:
        raise KeyError
    return entry[0]


def set_vcs_fields_to_salsa_user(control, salsa_user):
    old_vcs_url = control.source.get("Vcs-Git")
    control.source["Vcs-Git"] = "https://salsa.debian.org/{}/{}.git".format(
        salsa_user,
        control.source['Source']
    )
    new_vcs_url = control.source["Vcs-Git"]
    control.source["Vcs-Browser"] = "https://salsa.debian.org/{}/{}".format(
        salsa_user,
        control.source['Source']
    )
    return (old_vcs_url, new_vcs_url)


def set_maintainer_to_qa_team(control):
    if (QA_MAINTAINER == control.source.get('Maintainer') and
            'Uploaders' not in control.source):
        return False
    control.source["Maintainer"] = QA_MAINTAINER
    try:
        del control.source["Uploaders"]
    except KeyError:
        pass
    return True


class NoWnppBug(Exception):
    """No wnpp bug exists."""

    def __init__(self, package):
        self.package = package


class AlreadyOrphaned(Exception):
    """Package is already orphaned."""


class MissingControlFile(Exception):
    """Control file can not be found."""


def orphan(
        local_tree, subpath, update_changelog, committer, update_vcs=True,
        salsa_push=True, salsa_user="debian", dry_run=False,
        check_wnpp=True) -> OrphanResult:
    control_path = local_tree.abspath(
        osutils.pathjoin(subpath, "debian/control"))
    changelog_entries = []
    with ExitStack() as es:
        try:
            control = es.enter_context(ControlEditor(path=control_path))
        except FileNotFoundError as e:
            raise MissingControlFile(e.filename)
        if check_wnpp:
            try:
                wnpp_bug = find_wnpp_bug(control.source["Source"])
            except KeyError:
                raise NoWnppBug(control.source['Source'])
        else:
            wnpp_bug = None
        if set_maintainer_to_qa_team(control):
            if wnpp_bug is not None:
                changelog_entries.append(
                    "Orphan package - see bug %d." % wnpp_bug)
            else:
                changelog_entries.append("Orphan package.")
        result = OrphanResult(
            wnpp_bug=wnpp_bug, package=control.source["Source"])

        if update_vcs:
            (result.old_vcs_url,
             result.new_vcs_url) = set_vcs_fields_to_salsa_user(
                control, salsa_user)
            result.salsa_user = salsa_user
            if result.old_vcs_url == result.new_vcs_url:
                result.old_vcs_url = result.new_vcs_url = None
            else:
                changelog_entries.append(
                    "Update VCS URLs to point to Debian group."
                )
    if not changelog_entries:
        raise AlreadyOrphaned()
    if update_changelog in (True, None):
        cl_path = osutils.pathjoin(subpath, "debian/changelog")
        with ChangelogEditor(path=local_tree.abspath(cl_path)) as ce:
            ce.add_entry(changelog_entries)

    local_tree.commit(
        "Move package to QA team.", committer=committer, allow_pointless=False
    )

    result.pushed = False
    if update_vcs and salsa_push and result.new_vcs_url:
        parent_branch_url = local_tree.branch.get_parent()
        if parent_branch_url is not None:
            parent_branch = Branch.open(parent_branch_url)
        else:
            parent_branch = local_tree.branch
        push_result = push_to_salsa(
            local_tree,
            parent_branch,
            salsa_user,
            result.package,
            dry_run=dry_run,
        )
        if push_result:
            result.pushed = True
    return result


def move_instructions(package_name, salsa_user, old_vcs_url, new_vcs_url):
    yield "Please move the repository from {} to {}.".format(
        old_vcs_url, new_vcs_url)
    if urlparse(old_vcs_url).hostname == "salsa.debian.org":
        path = urlparse(old_vcs_url).path
        if path.endswith(".git"):
            path = path[:-4]
        yield "If you have the salsa(1) tool installed, run: "
        yield ""
        yield "    salsa fork --group={} {}".format(salsa_user, path)
    else:
        yield "If you have the salsa(1) tool installed, run: "
        yield ""
        yield "    git clone {} {}".format(old_vcs_url, package_name)
        yield "    salsa --group={} push_repo {}".format(
            salsa_user, package_name)


def report_fatal(code, description):
    if os.environ.get('SVP_API') == '1':
        with open(os.environ['SVP_RESULT'], 'w') as f:
            json.dump({
                'versions': versions_dict(),
                'result_code': code,
                'description': description}, f)
    logging.info('%s', description)


def main(argv=None):
    import argparse

    parser = argparse.ArgumentParser(prog="deb-move-orphaned")
    parser.add_argument(
        "--dry-run",
        help="Create branches but don't push or propose anything.",
        action="store_true",
        default=False,
    )
    parser.add_argument("--directory", type=str, help="Directory to open")
    parser.add_argument("--committer", help="Committer identity", type=str)
    parser.add_argument(
        "--no-update-changelog",
        action="store_false",
        default=True,
        dest="update_changelog",
        help="do not update the changelog",
    )
    parser.add_argument(
        "--update-changelog",
        action="store_true",
        dest="update_changelog",
        help="force updating of the changelog",
        default=None,
    )
    parser.add_argument(
        "--no-update-vcs",
        action="store_true",
        help="Do not move the VCS repository to the Debian team on Salsa.",
    )
    parser.add_argument(
        "--salsa-user",
        type=str,
        default="debian",
        help="Salsa user to push repository to.",
    )
    parser.add_argument(
        "--just-update-headers",
        action="store_true",
        help="Update the VCS-* headers, but don't actually "
        "clone the repository.",
    )
    parser.add_argument(
        "--no-check-wnpp", action="store_true",
        help="Do not check for WNPP bug.")
    args = parser.parse_args(argv)

    logging.basicConfig(format='%(message)s', level=logging.INFO)

    update_changelog = args.update_changelog
    if os.environ.get('DEB_UPDATE_CHANGELOG') == 'leave':
        update_changelog = False
    elif os.environ.get('DEB_UPDATE_CHANGELOG') == 'update':
        update_changelog = True

    tree, subpath = WorkingTree.open_containing(args.directory)

    try:
        result = orphan(
            tree,
            subpath=subpath,
            update_changelog=update_changelog,
            committer=args.committer,
            update_vcs=not args.no_update_vcs,
            dry_run=args.dry_run,
            salsa_user=args.salsa_user,
            salsa_push=not args.just_update_headers,
            check_wnpp=not args.no_check_wnpp,
        )
    except AlreadyOrphaned:
        report_fatal("nothing-to-do", "Already orphaned")
        return 0
    except NoWnppBug as e:
        report_fatal(
            "nothing-to-do",
            "Package %s is purported to be orphaned, "
            "but no open wnpp bug exists." % e.package,
        )
        return 1
    except FormattingUnpreservable as e:
        report_fatal(
            "formatting-unpreservable",
            "unable to preserve formatting while editing %s" % e.path)
        if hasattr(e, 'diff'):
            sys.stderr.writelines(e.diff())
        return 1
    except (ChangeConflict, GeneratedFile) as e:
        report_fatal(
            "generated-file", "unable to edit generated file: %r" % e
        )
        return 1
    except MissingControlFile as e:
        report_fatal("missing-control-file", "Missing control file: %r" % e)
        return 1

    if result.new_vcs_url:
        target_branch_url = vcs_git_url_to_bzr_url(result.new_vcs_url)
    else:
        target_branch_url = None

    if os.environ.get('SVP_API') == '1':
        with open(os.environ['SVP_RESULT'], 'w') as f:
            json.dump({
                'description': "Move package to QA team.",
                'versions': versions_dict(),
                'target-branch-url': target_branch_url,
                'value': 60,
                'context': result.json()}, f)

    if result.new_vcs_url:
        for line in move_instructions(
            result.package,
            result.salsa_user,
            result.old_vcs_url,
            result.new_vcs_url,
        ):
            logging.info("%s", line)

    return 0


if __name__ == '__main__':
    sys.exit(main(sys.argv[1:]))
